import collections
from pathlib import Path
import re

from mocodo.tools.various import first_missing_positive

from ..mocodo_error import MocodoError
from ..version_number import version


def set_defaults(template):
    result = {
        "transform_attribute": [],
        "transform_label": [],
        "transform_title": [],
        "transform_datatype": [],
        "transform_optionality": [],
        "label_role_separator": " ",
        "compose_primary_key": "_{label}_",
        "compose_normal_attribute": "{label}",
        "compose_foreign_key": "#{label}",
        "compose_primary_foreign_key": "_#{label}_",
        "add_unicity_constraints": [],
        "add_optionality_constraints": [],
        "transform_relation_name": [],
        "column_separator": ", ",
        "compose_relation": "{this_relation_name} ({columns})",
        "deleted_relation_separator": "",
        "compose_deleted_relation": "",
        "compose_deleted_relations": "",
        "transform_forced_relation": [],
        "transform_relation": [],
        "relation_separator": "\n",
        "compose_relational_schema": "{relations}",
        "transform_relational_schema": [],
    }
    result.update(template)
    result.setdefault("compose_ex_foreign_key", result["compose_normal_attribute"])
    result.setdefault("compose_primary_ex_foreign_key", result["compose_primary_key"])
    result.setdefault("compose_association_attribute", result["compose_normal_attribute"])
    result.setdefault("compose_association_primary_key", result["compose_primary_key"])
    result.setdefault("compose_deleted_child_attribute", result["compose_normal_attribute"])
    result.setdefault("compose_deleted_child_discriminator_", result["compose_normal_attribute"])
    result.setdefault("compose_deleted_child_discriminator_T", result["compose_normal_attribute"])
    result.setdefault("compose_deleted_child_discriminator_X", result["compose_normal_attribute"])
    result.setdefault("compose_deleted_child_discriminator_XT", result["compose_normal_attribute"])
    result.setdefault("compose_deleted_child_entity_name", result["compose_normal_attribute"])
    result.setdefault("compose_deleted_child_foreign_key", result["compose_foreign_key"])
    result.setdefault("compose_deleted_parent_discriminator_", result["compose_normal_attribute"])
    result.setdefault("compose_deleted_parent_discriminator_T", result["compose_normal_attribute"])
    result.setdefault("compose_deleted_parent_discriminator_X", result["compose_normal_attribute"])
    result.setdefault("compose_deleted_parent_discriminator_XT", result["compose_normal_attribute"])
    result.setdefault("compose_deleted_parent_attribute", result["compose_normal_attribute"])
    result.setdefault("compose_deleted_parent_foreign_key", result["compose_foreign_key"])
    result.setdefault("compose_deleted_parent_primary_key", result["compose_primary_key"])
    result.setdefault("compose_stopped_foreign_key", result["compose_foreign_key"])
    result.setdefault("compose_outer_attribute", result["compose_normal_attribute"])
    result.setdefault("compose_outer_primary_key", result["compose_primary_key"])
    result.setdefault("compose_parent_primary_key", result["compose_primary_foreign_key"])
    result.setdefault("compose_strengthening_primary_foreign_key", result["compose_primary_foreign_key"])
    result.setdefault("compose_strengthening_primary_ex_foreign_key", result["compose_primary_key"])
    result.setdefault("compose_unsourced_foreign_key", result["compose_normal_attribute"])
    result.setdefault("compose_unsourced_primary_foreign_key", result["compose_primary_key"])
    result.setdefault("compose_ex_foreign_key", result["compose_normal_attribute"])
    result.setdefault("compose_primary_ex_foreign_key", result["compose_primary_key"])
    result.setdefault("compose_deleted_child_ex_foreign_key", result["compose_normal_attribute"])
    result.setdefault("compose_deleted_parent_ex_foreign_key", result["compose_normal_attribute"])
    result.setdefault("compose_stopped_ex_foreign_key", result["compose_normal_attribute"])
    result.setdefault("compose_unsourced_ex_foreign_key", result["compose_normal_attribute"])
    result.setdefault("compose_unsourced_primary_ex_foreign_key", result["compose_primary_key"])
    return result


class Relations:

    def __init__(self, mcd, params):
        
        self.mcd = mcd
        self.output_stem = Path(params["output_name"]).stem
        self.ensure_no_reciprocical_relative_entities()
        self.freeze_strengthening_foreign_key_migration = set()
        self.relations = {}
        self.relations_from_entities()
        self.strengthen_weak_identifiers()
        self.inheritance_parent_or_children_to_delete = self.find_inheritance_parent_or_children_to_delete()
        self.strengthen_children()
        self.strengthen_parents()
        self.process_associations()
        self.process_inheritances()
        self.delete_inheritance_parent_or_children_to_delete()
        self.delete_deletable_relations()
        self.make_primary_keys_first()
        self.relations = dict(sorted(self.relations.items()))

    
    def get_text(self, template):
        
        def transform(string, transformation):
            for d in template[transformation]:
                while True:
                    try:
                        (string, n) = re.subn(d["search"], d["replace"], str(string))
                    except:
                        raise MocodoError(27, _('Cannot compile the regular expression "{regex}" or the remplacement string "{replace}" in a relation template producing "*{stem_suffix}.{extension}" files.').format(regex=d["search"], replace=d["replace"], stem_suffix=template["stem_suffix"], extension=template["extension"])) # fmt: skip
                    if n == 0 or not d.get("iterated"):
                        break
            return string
        
        template = set_defaults(template)
        
        def make_non_disambiguated_labels_from_attributes():
            for relation in self.relations.values():
                for column in relation["columns"]:
                    column["non_disambiguated_label"] = transform(column["attribute"], "transform_attribute")
        make_non_disambiguated_labels_from_attributes()

        def make_labels_from_non_disambiguated_labels():
            for relation in self.relations.values():
                for column in relation["columns"]:
                    if not column["leg_note"]:
                        column["label"] = column["non_disambiguated_label"]
                    elif column["leg_note"].startswith("-"):
                        column["label"] = column["leg_note"][1:]
                    elif column["leg_note"].startswith("+"):
                        column["label"] = column["non_disambiguated_label"] + column["leg_note"][1:]
                    elif " " in column["leg_note"]:
                        column["label"] = column["non_disambiguated_label"]
                    else:
                        column["label"] = column["non_disambiguated_label"] + template["label_role_separator"] + column["leg_note"]
            # After labels have been disambiguated by roles, ensure all of them are distinct.
            for relation in self.relations.values():
                occurrences = collections.Counter(column["label"] for column in relation["columns"])
                occurrences = dict(c for c in occurrences.items() if c[1] > 1)
                for column in reversed(relation["columns"]):
                    if column["label"] in occurrences and column["nature"] != "primary_key":
                        occurrences[column["label"]] -= 1
                        column["label"] = column["label"] + template["label_role_separator"] + str(occurrences[column["label"]] + 1)
        make_labels_from_non_disambiguated_labels()

        def transform_labels(): # after labels have been disambiguated by roles
            for relation in self.relations.values():
                for column in relation["columns"]:
                    column["label"] = transform(column["label"], "transform_label")
        transform_labels()
    
        def add_fillers():
            for relation in self.relations.values():
                label_max_length = max(len(column["label"]) for column in relation["columns"])
                for column in relation["columns"]:
                    column["filler"] = " " * (label_max_length - len(column["label"]) + 1)
        add_fillers()

        data = {}
        data["stem"] = self.output_stem
        data["title"] = transform(self.mcd.title, "transform_title")
        data["version"] = version
        lines = []
        for (__, relation) in sorted(self.relations.items()): # For the double underscore, see __main__.py
            data["this_relation_name"] = transform(relation["this_relation_name"], "transform_relation_name")
            data["is_forced"] = relation["is_forced"]
            fields = []
            for column in relation["columns"]:
                column["datatype"] = transform(column["datatype"], "transform_datatype")
                column["optionality"] = transform(column["optionality"], "transform_optionality")
                data.update(column)
                field = template["compose_%s" % column["nature"]].format(**data)
                if column["unicities"]:
                    field = transform(field, "add_unicity_constraints").format(**data)
                if column["optionality"]:
                    field = transform(field, "add_optionality_constraints").format(**data)
                fields.append(field)
            data["columns"] = template["column_separator"].join(fields)
            line = template["compose_relation"].format(**data)
            if relation["is_forced"]:
                line = transform(line, "transform_forced_relation")
            line = transform(line, "transform_relation")
            lines.append(line)
        if template.get("stem_suffix") == "_mld" and template.get("extension") == "mcd": # relational diagram
            lines = self.map_mcd_layout_onto_mld(lines)
        data["relations"] = template["relation_separator"].join(lines)

        if self.deleted_relations:
            lines = []
            for deleted_relation in self.deleted_relations:
                lines.append(template["compose_deleted_relation"].format(this_relation_name=deleted_relation))
            deleted_relation_lines = template["deleted_relation_separator"].join(lines)
            data["deleted_relations"] = template["compose_deleted_relations"].format(deleted_relation_lines=deleted_relation_lines)
        else:
            data["deleted_relations"] = ""

        data["relations"] = template["compose_relational_schema"].format(**data)
        result = transform(data["relations"], "transform_relational_schema")
        return result


    # private

    def map_mcd_layout_onto_mld(self, lines):
        rows = [[]]
        all_commas = [True] * self.mcd.col_count
        for row in self.mcd.rows:
            for (i, box) in enumerate(row):
                for line in lines:
                    if line.startswith(box.name_view + ":"):
                        rows[-1].append(line)
                        all_commas[i] = False
                        break
                else:
                    rows[-1].append(":")
            rows.append([])
        rows.pop()
        for row in rows:
            row[:] = [x for (x, all_comma) in zip(row, all_commas) if not all_comma]
        lines = []
        for row in rows:
            lines.append(":")
            for x in row:
                lines.append(x)
                lines.append(":")
            lines.append("\n")
        lines.pop()
        return lines

    def may_retrieve_distant_leg_note(self, leg, attribute):
        if leg.entity_bid in self.inheritance_parent_or_children_to_delete:
            for d in self.relations[leg.entity.bid]["columns"]:
                if d["attribute"] == attribute["attribute"] and d["adjacent_source"] == attribute["adjacent_source"]:
                    return d["leg_note"]
        return leg.note

    def may_retrieve_distant_outer_source(self, leg, attribute):
        if leg.entity_bid in self.inheritance_parent_or_children_to_delete:
            for d in self.relations[leg.entity_bid]["columns"]:
                if d["attribute"] == attribute["attribute"] and d["adjacent_source"] == attribute["adjacent_source"]:
                    return d["outer_source"]
        return leg.entity.name_view

    def ensure_no_reciprocical_relative_entities(self):
        for association in self.mcd.associations.values():
            if association.is_invisible:
                continue
            weak_count = 0
            for leg in association.legs:
                if leg.kind == "strengthening":
                    weak_count += 1
                    if weak_count == 2:
                        raise MocodoError(11, _('Reciprocal relative identification around {association}.').format(association=association.bid)) # fmt: skip

    @staticmethod
    def pop_optionality_from_datatype(attribute, is_primary = False):
        (datatype, n) = re.subn(r"(?i) *\bnot +null\b *", " ", attribute.datatype)
        attribute.datatype = datatype.strip() # suppress the NOT NULL, if any
        if n: # A NOT NULL has been suppressed
            return "!" # the attribute is mandatory (as explicitely stated)
        (datatype, n) = re.subn(r"(?i) *\bnull\b *", " ", attribute.datatype)
        attribute.datatype = datatype.strip() # suppress the NULL, if any
        if is_primary:
            return "!" # the attribute is mandatory (even if not explicitely stated)
        if n: # A NULL has been suppressed
            return "?" # the attribute is optional (as explicitely stated)
        return "" # the attribute is neither optional nor mandatory

    def relations_from_entities(self):
        for (name, entity) in self.mcd.entities.items():
            if entity.is_invisible:
                continue
            self.relations[name] = {
                "this_relation_name": entity.name_view,
                "is_forced": False, # an entity naturally results in a relation. No need to force it.
                "is_protected": entity.is_protected,
                "columns": [],
                "existing_unicity_numbers": set(),
            }
            for attribute in entity.attributes:
                if attribute.label.strip() == "":
                    continue # ignore empty attributes
                nature = "primary_key" if attribute.kind in ("strong", "weak") else "normal_attribute"
                unicities = "".join(c for c in sorted(attribute.id_groups) if c != "0")
                is_primary = (nature == "primary_key")
                optionality = self.pop_optionality_from_datatype(attribute, is_primary)
                self.relations[name]["columns"].append({
                    "attribute": attribute.label,
                    "optionality": optionality,
                    "datatype": attribute.datatype,
                    "adjacent_source": None,
                    "outer_source": None,
                    "association_name": None,
                    "leg_note": None,
                    "is_primary": is_primary,
                    "nature": nature,
                    "unicities": unicities,
                })
                self.relations[name]["existing_unicity_numbers"].update(map(int, unicities))

    def strengthen_weak_identifiers(self):
        for entity in self.mcd.entities.values():
            entity.is_strong_or_strengthened = not entity.strengthening_legs
        remaining_entities = [entity for entity in self.mcd.entities.values() if not entity.is_strong_or_strengthened and not entity.is_invisible]
        while remaining_entities:
            for entity in remaining_entities:
                strengthening_entities_via_associations = []
                for strengthening_leg in entity.strengthening_legs:
                    association = strengthening_leg.association
                    for other_leg in association.legs:
                        other_entity = other_leg.entity
                        if other_entity == entity:
                            continue
                        if not other_entity.is_strong_or_strengthened:
                            break # weak entity linked to a weak entity
                        strengthening_entities_via_associations.append((other_entity, association))
                    else:
                        continue
                    break
                if strengthening_entities_via_associations:
                    for (strengthening_entity, association) in strengthening_entities_via_associations:
                        # find the potential note on the strengthening leg
                        for leg in association.legs:
                            if leg.entity_bid == strengthening_entity.bid:
                                leg_note = leg.note
                                break
                        else:
                            leg_note = None
                        # migrate the whole primary key of the strengthening entity into the weak one
                        self.relations[entity.bid]["columns"][0:0] = [{
                                "attribute": attribute["attribute"],
                                "optionality": "!",
                                "datatype": attribute["datatype"],
                                "adjacent_source": strengthening_entity.name_view,
                                "outer_source": strengthening_entity.name_view,
                                "association_name": association.name_view,
                                "leg_note": leg_note,
                                "is_primary": True,
                                "nature": "strengthening_primary_foreign_key",
                                "unicities": "",
                            } for attribute in self.relations[strengthening_entity.bid]["columns"] if attribute["is_primary"]]
                        self.freeze_strengthening_foreign_key_migration.add((entity.bid, association.bid, strengthening_entity.bid))
                    remaining_entities.remove(entity)
                    entity.is_strong_or_strengthened = True
                    break
            else:
                if len(remaining_entities) == 1:
                    raise MocodoError(16, _('A weak entity (here, {entity}) cannot be strengthened by itself.').format(entity=remaining_entities[0].bid)) # fmt: skip
                else:
                    remaining_entity_names = ", ".join('"%s"' % entity.raw_name for entity in remaining_entities)
                    raise MocodoError(17, _('Cycle of weak entities in {entities}.').format(entities=remaining_entity_names)) # fmt: skip

    def find_inheritance_parent_or_children_to_delete(self):
        result = set()
        for inheritance in self.mcd.inheritances:
            parent_leg = inheritance.legs[0]
            if inheritance.kind in ("<-", "<="):
                for child_leg in inheritance.legs[1:]:
                    result.add(child_leg.entity_bid)
            elif inheritance.kind == "=>":
                if "T" not in inheritance.name_view:
                    raise MocodoError(25, _('Totality (/T\\ or /XT\\) is mandatory for "=>" inheritance of parent "{name}".').format(name=parent_leg.entity_raw_name)) # fmt: skip
                result.add(parent_leg.entity_bid)
        return result

    def strengthen_children(self):
        """
        Migrate the parent's identifier in the children. This is obviously necessary if the parent
        must disappear (`=>` + totality) or its attributes copied in its children. Otherwise, the
        children must disappear. But in the case they are connected to other entities, they need
        an identifier to make possible to apply the remaining rules.
        """
        for inheritance in self.mcd.inheritances:
            parent_leg = inheritance.legs[0]
            to_be_deleted = parent_leg.entity_bid in self.inheritance_parent_or_children_to_delete
            for child_leg in inheritance.legs[1:]: 
                self.relations[child_leg.entity_bid]["columns"][0:0] = [{
                    "attribute": attribute["attribute"],
                    "optionality": "!",
                    "datatype": attribute["datatype"],
                    "adjacent_source": parent_leg.entity.name_view,
                    "outer_source": parent_leg.entity.name_view,
                    "association_name": inheritance.name_view,
                    "leg_note": None,
                    "is_primary": True,
                    "nature": "deleted_parent_primary_key" if to_be_deleted else "parent_primary_key",
                    "unicities": "",
                } for attribute in self.relations[parent_leg.entity_bid]["columns"] if attribute["is_primary"]]

    def strengthen_parents(self):
        """
        Migrate the optional children's discriminators in their parent when it disappears (totality + =>).
        In this case, this discriminator should further migrate with the identifier of the parent.
        """
        for inheritance in self.mcd.inheritances:
            if inheritance.kind == "=>":
                parent_leg = inheritance.legs[0]
                self.relations[parent_leg.entity_bid]["columns"].extend({ 
                    "attribute": attribute.label,
                    "optionality": "!" if 'T' in inheritance.name_view else "?",
                    "datatype": attribute.datatype or "UNSIGNED_INT_PLACEHOLDER",
                    "adjacent_source": None,
                    "outer_source": None,
                    "association_name": inheritance.name_view,
                    "leg_note": None,
                    "is_primary": False,
                    "nature": f"deleted_parent_discriminator_{inheritance.name_view}",
                    "unicities": "",
                } for attribute in inheritance.attributes)

    def process_associations(self):
        for association in self.mcd.associations.values():
            if association.is_invisible:
                continue
            is_cluster = (association.kind == "cluster")
            df_leg = None
            for leg in association.legs:
                if leg.entity.is_invisible:
                    continue
                if leg.card[1] == "1":
                    df_leg = leg
                    if leg.card[0] == "1":
                        break # elect the first leg with cardinality 11
            
            if df_leg is None or association.is_protected:
                # make a relation of this association
                self.relations[association.bid] = {
                    "this_relation_name": association.name_view,
                    "is_forced": bool(df_leg), # if this association has a 11 leg, being here means it is protected: it must be forced into a relation
                    "columns": [],
                    "existing_unicity_numbers": set(),
                }
                for leg in association.legs:
                    if leg.entity.is_invisible:
                        continue
                    for attribute in self.relations[leg.entity_bid]["columns"]:
                        if attribute["is_primary"]:
                            outer_source = self.may_retrieve_distant_outer_source(leg, attribute)
                            if is_cluster:
                                self.relations[association.bid]["columns"].append({ # gather all migrant attributes
                                    "attribute": attribute["attribute"],
                                    "optionality": "!",
                                    "datatype": attribute["datatype"],
                                    "adjacent_source": leg.entity.name_view,
                                    "outer_source": outer_source,
                                    "association_name": association.name_view,
                                    "leg_note": leg.note,
                                    "is_primary": leg.is_in_elected_group,
                                    "nature": "primary_foreign_key" if leg.is_in_elected_group else "stopped_foreign_key",
                                    "unicities": leg.unicities,
                                })
                                self.relations[association.bid]["existing_unicity_numbers"].update(map(int, leg.unicities))
                            elif association.is_protected and df_leg is not None and leg is not df_leg:
                                self.relations[association.bid]["columns"].append({ # gather all migrant attributes
                                    "attribute": attribute["attribute"],
                                    "optionality": "!",
                                    "datatype": attribute["datatype"],
                                    "adjacent_source": leg.entity.name_view,
                                    "outer_source": outer_source,
                                    "association_name": association.name_view,
                                    "leg_note": leg.note,
                                    "is_primary": False,
                                    "nature": "stopped_foreign_key",
                                    "unicities": "",
                                })
                            else:
                                self.relations[association.bid]["columns"].append({ # gather all migrant attributes
                                    "attribute": attribute["attribute"],
                                    "optionality": "!",
                                    "datatype": attribute["datatype"],
                                    "adjacent_source": leg.entity.name_view,
                                    "outer_source": outer_source,
                                    "association_name": association.name_view,
                                    "leg_note": leg.note,
                                    "is_primary": True,
                                    "nature": "unsourced_primary_foreign_key" if outer_source is None else "primary_foreign_key",
                                    "unicities": "",
                                })
                        elif attribute["nature"].startswith("deleted_parent_discriminator"):
                            self.relations[association.bid]["columns"].append({
                                "attribute": attribute["attribute"],
                                "optionality": "!",
                                "datatype": attribute["datatype"],
                                "adjacent_source": leg.entity.name_view,
                                "outer_source": None,
                                "association_name": association.name_view,
                                "leg_note": leg.note,
                                "is_primary": False,
                                "nature": attribute["nature"],
                                "unicities": "",
                            })
                self.relations[association.bid]["columns"].extend({ # and the attributes already existing in the association
                        "attribute": attribute.label,
                        "optionality": "!" if attribute.kind == "strong" else self.pop_optionality_from_datatype(attribute),
                        "datatype": attribute.datatype,
                        "adjacent_source": None,
                        "outer_source": None,
                        "association_name": association.name_view,
                        "leg_note": None,
                        "is_primary": attribute.kind == "strong",
                        "nature": "association_primary_key" if attribute.kind == "strong" else "association_attribute",
                        "unicities": "",
                    } for attribute in association.attributes if attribute.label.strip() != ""
                )
                continue

            # No relation will be created from this association.

            # Check the number of /?1 legs (pegs)
            df_pegs = []
            for leg in association.legs:
                if leg.entity.is_invisible:
                    continue
                if leg.card[1] == "1" and leg.kind == "cluster_peg":
                    df_pegs.append(leg)
            
            if df_pegs: # A cluster with at least one /?1 leg (peg)
                for df_peg in df_pegs:
                    # All migrating attribute must belong to the same new unicity group
                    unicities = str(first_missing_positive(self.relations[df_peg.entity_bid]["existing_unicity_numbers"]))
                    self.relations[df_leg.entity_bid]["existing_unicity_numbers"].update(map(int, unicities))
                    for leg in association.legs:
                        if leg is df_peg:
                            continue
                        for attribute in list(self.relations[leg.entity_bid]["columns"]): # traverse a copy...
                            # ... to prevent an infinite migration of the child discriminator
                            if attribute["is_primary"]:
                                # Their primary keys must migrate in `entity_name`.
                                outer_source = self.may_retrieve_distant_outer_source(leg, attribute)
                                self.relations[df_peg.entity_bid]["columns"].append({
                                    "attribute": attribute["attribute"],
                                    "optionality": "!",
                                    "datatype": attribute["datatype"],
                                    "adjacent_source": leg.entity.name_view,
                                    "outer_source": outer_source,
                                    "association_name": association.name_view,
                                    "leg_note": leg.note,
                                    "is_primary": False,
                                    "nature": "unsourced_foreign_key" if outer_source is None else "foreign_key",
                                    "unicities": unicities,
                                    # NB: technically, an unsourced foreign key is not foreign anymore
                                })
                # The association attributes must migrate only once, df_leg is elected.
                # Delay the addition of the association attributes after the else part.

            else:
                # A normal DF association. Traverse the other legs to find the attributes to migrate.
                for leg in association.legs:
                    if leg.entity.is_invisible:
                        continue
                    if leg is not df_leg:
                        if (df_leg.entity_bid, association.bid, leg.entity_bid) not in self.freeze_strengthening_foreign_key_migration:
                            unicities = ""
                            if leg.card[1] == "1": # *1 --(DF)-- 11 => the migrating attribute must be made unique, find a new unicity group
                                unicities = str(first_missing_positive(self.relations[df_leg.entity_bid]["existing_unicity_numbers"]))
                            for attribute in list(self.relations[leg.entity_bid]["columns"]): # traverse a copy...
                                # ... to prevent an infinite migration of the child discriminator
                                optionality = "!" if df_leg.card[0] == "1" else "?"
                                if attribute["is_primary"]:
                                    # Their primary keys must migrate in df_leg.entity_bid.
                                    outer_source = self.may_retrieve_distant_outer_source(leg, attribute)
                                    self.relations[df_leg.entity_bid]["columns"].append({
                                        "attribute": attribute["attribute"],
                                        "optionality": optionality,
                                        "datatype": attribute["datatype"],
                                        "adjacent_source": leg.entity.name_view,
                                        "outer_source": outer_source,
                                        "association_name": association.name_view,
                                        "leg_note": leg.note,
                                        "is_primary": False,
                                        "nature": "unsourced_foreign_key" if outer_source is None else "foreign_key",
                                        "unicities": unicities,
                                        # NB: technically, an unsourced foreign key is not foreign anymore
                                    })
                                    self.relations[df_leg.entity_bid]["existing_unicity_numbers"].update(map(int, unicities))
                                elif attribute["nature"].startswith("deleted_parent_discriminator"):
                                    self.relations[df_leg.entity_bid]["columns"].append({
                                        "attribute": attribute["attribute"],
                                        "optionality": optionality,
                                        "datatype": attribute["datatype"],
                                        "adjacent_source": leg.entity.name_view,
                                        "outer_source": None,
                                        "association_name": association.name_view,
                                        "leg_note": leg.note,
                                        "is_primary": False,
                                        "nature": attribute["nature"],
                                        "unicities": "",
                                    })
                
            # Add the attributes already existing in the association
            self.relations[df_leg.entity_bid]["columns"].extend([{
                    "attribute": attribute.label,
                    "optionality": "!" if attribute.kind == "strong" else self.pop_optionality_from_datatype(attribute),
                    "datatype": attribute.datatype,
                    "association_name": association.name_view,
                    "adjacent_source": None,
                    "outer_source": None,
                    "leg_note": None,
                    "is_primary": attribute.kind == "strong",
                    "nature": "outer_primary_key" if attribute.kind == "strong" else "outer_attribute",
                    "unicities": "",
                } for attribute in association.attributes if attribute.label.strip() != ""])


    def process_inheritances(self):
        for inheritance in self.mcd.inheritances:
            parent_leg = inheritance.legs[0]
            if inheritance.kind == "=>": # total migration: parent > children
                for child_leg in inheritance.legs[1:]: 
                    # migrate the parent's attributes, except those of nature "deleted_parent_discriminator"
                    self.relations[child_leg.entity_bid]["columns"][0:0] = [{
                        "attribute": attribute["attribute"],
                        "optionality": attribute["optionality"],
                        "datatype": attribute["datatype"],
                        "adjacent_source": parent_leg.entity.name_view,
                        "outer_source": self.may_retrieve_distant_outer_source(parent_leg, attribute),
                        "association_name": inheritance.name_view,
                        "leg_note": self.may_retrieve_distant_leg_note(parent_leg, attribute),
                        "is_primary": False,
                        "nature": "deleted_parent_foreign_key" if attribute["nature"] == "foreign_key" else "deleted_parent_attribute",
                        "unicities": "",
                    } for attribute in self.relations[parent_leg.entity_bid]["columns"] if not attribute["is_primary"] and not attribute["nature"].startswith("deleted_parent_discriminator")]
            else: # migration: triangle attributes > parent
                self.relations[parent_leg.entity_bid]["columns"].extend({ 
                    "attribute": attribute.label,
                    "optionality": "!" if 'T' in inheritance.name_view else "?",
                    "datatype": attribute.datatype or "UNSIGNED_INT_PLACEHOLDER",
                    "adjacent_source": None,
                    "outer_source": None,
                    "association_name": inheritance.name_view,
                    "leg_note": None,
                    "is_primary": False,
                    "nature": f"deleted_child_discriminator_{inheritance.name_view}", # "", "X", "T" or "XT"
                    "unicities": "",
                } for attribute in inheritance.attributes)
                if inheritance.kind in ("<-", "<="): # migration: children > parent, and suppress children
                    for child_leg in inheritance.legs[1:]:
                        if inheritance.kind == "<=":
                            # make the child's name a boolean attribute of the parent
                            self.relations[parent_leg.entity_bid]["columns"].append({
                                "attribute": _('is {name}').format(name=child_leg.entity_bid.lower()),
                                "optionality": "!",
                                "datatype": "BOOLEAN_PLACEHOLDER",
                                "adjacent_source": child_leg.entity.name_view,
                                "outer_source": child_leg.entity.name_view,
                                "association_name": inheritance.name_view,
                                "leg_note": None,
                                "is_primary": False,
                                "nature": "deleted_child_entity_name",
                                "unicities": "",
                            })
                        # migrate all child's attributes
                        for attribute in self.relations[child_leg.entity_bid]["columns"]:
                            if attribute["nature"].endswith("parent_primary_key"):
                                continue # except the "strengthening" parent identifier
                            self.relations[parent_leg.entity_bid]["columns"].append({
                                "attribute": attribute["attribute"],
                                "optionality": "?",
                                "datatype": attribute["datatype"],
                                "adjacent_source": child_leg.entity.name_view,
                                "outer_source": self.may_retrieve_distant_outer_source(child_leg, attribute),
                                "association_name": inheritance.name_view,
                                "leg_note": self.may_retrieve_distant_leg_note(child_leg, attribute),
                                "is_primary": False,
                                "nature": "deleted_child_foreign_key" if attribute["nature"] == "foreign_key" else "deleted_child_attribute",
                                "unicities": "",
                            })
    
    def delete_inheritance_parent_or_children_to_delete(self):
        for entity_to_delete in self.inheritance_parent_or_children_to_delete:
            del self.relations[entity_to_delete]

    def delete_deletable_relations(self):
        deleted_outer_sources = set()
        for (bid, relation) in list(self.relations.items()):
            if not relation.get("is_protected") and all(column["nature"] == "primary_key" for column in relation["columns"]):
                del self.relations[bid]
                deleted_outer_sources.add(relation["this_relation_name"])
        for relation in self.relations.values():
            for column in relation["columns"]:
                if column["outer_source"] in deleted_outer_sources:
                    column["nature"] = column["nature"].replace("foreign", "ex_foreign")
        self.deleted_relations = sorted(deleted_outer_sources)

    def make_primary_keys_first(self):
        for relation in self.relations.values():
            relation["columns"].sort(key=lambda column: not column["is_primary"])



import argparse
import contextlib
import gettext
import json
import locale
import os
import random
import re
import importlib
import sys
from pathlib import Path

from mocodo.tools.string_tools import strip_surrounds
from mocodo.tools.various import invert_dict

from .version_number import version
from .mocodo_error import MocodoError

SCRIPT_DIRECTORY = os.path.dirname(os.path.realpath(os.path.join(__file__)))

class ArgumentDefaultsRawDescriptionHelpFormatter(
    argparse.ArgumentDefaultsHelpFormatter,
    argparse.RawDescriptionHelpFormatter,
):
    pass

def init_localization(language):
    if not language:
        if (
            sys.platform.lower().startswith("darwin")
            and os.system("defaults read -g AppleLanguages > /tmp/languages.txt") == 0
        ):
            language = re.search(r"\W*(\w+)", Path("/tmp/languages.txt").read_text(encoding="utf8")).group(1)
        else:
            try:
                language = locale.getdefaultlocale()[0][:2]
            except:
                language = "en"
    try:
        path = Path(f"{SCRIPT_DIRECTORY}/resources/i18n/messages_{language}.mo")
        with path.open("rb") as mo_contents:
            trans = gettext.GNUTranslations(mo_contents)
    except IOError:
        trans = gettext.NullTranslations()

    # Beware that the following line sets the global variable `_`. Using a single underscore
    # would clash with the gettext alias, EVEN IF THE CODE IS NOT REACHED: the compiler would
    # consider that _ is a local variable, and shadow the global one, resulting in:
    # UnboundLocalError: local variable '_' referenced before assignment
    # Solution : use `__` instead of `_` in any function that uses the gettext alias.
    trans.install()
    return language

class Transformations:

    def __init__(self, language):
        metadata_path = Path(f"{SCRIPT_DIRECTORY}/resources/transformations.json")
        self.metadata = json.loads(metadata_path.read_text(encoding="utf8"))
        index_path = Path(f"{SCRIPT_DIRECTORY}/resources/relation_templates/_index.json")
        template_data = json.loads(index_path.read_text(encoding="utf8"))
        for (stem, d) in template_data.items():
            if f"help_{language}" in d:
                d["help"] = d.pop(f"help_{language}")
                continue
            try:
                d["help"] = d[f"help_en"]
            except KeyError:
                raise MocodoError(22, _("The template '{stem}' doesn't have a help message in language of code '{language}' or 'en'.").format(stem=stem, language=language)) # fmt: skip
        self.metadata.update(template_data)
        self.normalize = invert_dict({k: v["aliases"] for (k, v) in self.metadata.items()})
        self.normalize.update((k, k) for k in self.metadata)
        # recreate the dictionary to have it in alphabetical order
        self.metadata = {k: self.metadata[k] for k in sorted(self.metadata.keys(), key=lambda x: x.lower())}
        self.operations = {"rw": [], "cv": []}
        self.op_tk_rewritings = set(k for (k, v) in self.metadata.items() if v.get("op_tk"))
        self.args_to_delete = ["-t", "-T", "--transform"]
        self.opt_to_restore = " "
    
    def extract_transformation_subargs(self, arg):
        (subopt, __, tail) = arg.partition(":")
        try:
            subopt = self.normalize[subopt.lower()]
        except:
            valid = ", ".join(sorted(self.normalize.keys()))
            raise MocodoError(45, _("The transformation '{subopt}' is not among the possible ones:\n{valid}.").format(subopt=subopt, valid=valid)) # fmt: skip
        result = extract_subargs(f"{subopt}:{tail}")
        category = self.metadata[subopt]["category"]
        self.operations[category].append(result)
        if category == "rw":
            self.args_to_delete.append(arg)
        else:
            self.opt_to_restore = " -t "
        return result


def extract_subargs(arg):
    (subopt, _, tail) = arg.partition(":")
    subargs = {}
    for string in filter(None, tail.split(",")):
        subsubopt, equal, subsubarg = string.partition("=")
        if equal: # subopt:subsubopt= / subopt:subsubopt=subsubarg
            subsubarg = strip_surrounds(subsubarg, "''")
            subsubarg = strip_surrounds(subsubarg, '""')
        else: # subopt:subsubopt
            subsubarg = None
        subargs[subsubopt] = subsubarg
    return (subopt, subargs)

def rate(string):
    try:
        value = float(string)
    except ValueError:
        msg = f"The rate {repr(string)} cannot be coerced to float"
        raise argparse.ArgumentTypeError(msg)
    if not (0 <= value <= 1):
        msg = f"The rate {repr(string)} is not between 0 and 1"
        raise argparse.ArgumentTypeError(msg)
    return value


def scale(string):
    try:
        value = float(string)
    except ValueError:
        msg = f"The scale {repr(string)} cannot be coerced to float"
        raise argparse.ArgumentTypeError(msg)
    if value <= 0:
        msg = f"The scale {repr(string)} is not strictly positive"
        raise argparse.ArgumentTypeError(msg)
    return value


def non_negative_integer(string):
    try:
        value = int(string)
    except ValueError:
        msg = f"The value {repr(string)} cannot be coerced to an integer"
        raise argparse.ArgumentTypeError(msg)
    if value < 0:
        msg = f"The integer {repr(string)} is negative"
        raise argparse.ArgumentTypeError(msg)
    return value


def positive_integer(string):
    try:
        value = int(string)
    except ValueError:
        msg = f"The value {repr(string)} cannot be coerced to an integer"
        raise argparse.ArgumentTypeError(msg)
    if value <= 0:
        msg = f"The integer {repr(string)} is negative or zero"
        raise argparse.ArgumentTypeError(msg)
    return value

def parsed_arguments(args):
    # When mocodo is called from the command line, `args` is just `sys.argv[1:]`.
    # When it is imported, `args` is a list of strings constructed with shlex.split().

    parser = argparse.ArgumentParser(
        prog="mocodo",
        add_help=False,
        formatter_class=ArgumentDefaultsRawDescriptionHelpFormatter,
        allow_abbrev=False,
    )

    mocodo_group = parser.add_argument_group()
    io_group = parser.add_argument_group()
    aspect_group = parser.add_argument_group()

    if sys.platform.lower().startswith("darwin"):
        default_params = {
            "encodings": ["utf8", "macroman"],
            "shapes": "copperplate",
        }
    elif sys.platform.lower().startswith("win"):
        default_params = {
            "encodings": ["utf8", "ISO_8859-15"],
            "shapes": "trebuchet",
        }
    else:  # linux
        default_params = {
            "encodings": ["utf8", "ISO_8859-15"],
            "shapes": "serif",
        }

    # First, add the arguments that need some preprocessing.
    # The help messages will be added later, after the localization.

    mocodo_group.add_argument("--language",
        metavar="CODE",
        type=str,
    )
    language_action = mocodo_group._group_actions[-1]
    io_group.add_argument("--params_path",
        metavar="PATH",
        default="params.json",
    )
    params_path_action = io_group._group_actions[-1]
    io_group.add_argument("--input", "-i",
        metavar="PATH",
        default=""
    )
    input_action = io_group._group_actions[-1]
    
    (args, remaining_args) = parser.parse_known_args(args)

    if not args.input:
        # no input file is specified, use the pristine sandbox
        source = Path(SCRIPT_DIRECTORY, "resources", "pristine_sandbox.mcd").read_text(encoding="utf8")
        args.input = "sandbox.mcd"
        Path(args.input).write_text(source, encoding="utf8")
    elif not args.input.endswith(".mcd"):
        # the input file has no extension, assume it is a stem
        args.input += ".mcd"
    default_params["input"] = args.input

    if os.path.exists(args.params_path):
        default_params.update(json.loads(Path(args.params_path).read_text(encoding="utf8")))
    if not default_params["input"]:
        default_params["input"] = "sandbox.mcd"
    default_params["language"] = init_localization(default_params.get("language", args.language))
    default_params.setdefault("output_dir", os.path.dirname(default_params["input"]))

    # Add localized messages.

    mocodo_group.title = _("OPTIONS ON MOCODO ITSELF")
    io_group.title = _("INPUT/OUTPUT")
    aspect_group.title = _("ASPECT OF THE GRAPHICAL OUTPUT")

    language_action.help = _("override the automatic localization of the messages with the given language code (e.g., 'fr', 'en', ...)")
    params_path_action.help = _("the path of the parameter file. If omitted, use 'params.json' in the input directory. If non existent, use default parameters.")
    input_action.help = _("the path of the input file. By default, the output files will be generated in the same directory")

    parser.description = re.sub(r"(?m)^(        )", "", _("""
        NAME:
        Mocodo - An Entity-Relation Diagram Generator.

        DESCRIPTION:
        Mocodo is an open-source tool for designing and teaching relational databases.
        It takes as an input a textual description of both entities and associations
        of an entity-relationship diagram (ERD). It outputs a vectorial drawing in SVG
        and a relational schema in various formats (SQL, LaTeX, Markdown, etc.).

        NOTE:
        Each one of the following values is:
        - explicitely specified by the user as a command line option;
        - otherwise, retrieved from a file located at --params_path;
        - otherwise, retrieved from a file named 'params.json' in the input directory;
        - otherwise, calculated from a default value, possibly dependant of your system.
    """)) # fmt: skip

    parser.epilog = re.sub(r"(?m)^(        )", "", _("""
        SEE ALSO:
          Online version        https://mocodo.net
          Source code           https://github.com/laowantong/mocodo
          Documentation         https://rawgit.com/laowantong/mocodo/master/doc/fr_refman.html
          Cheat sheet for -t    https://github.com/laowantong/mocodo/blob/master/doc/fr_cheat_sheet.md

        LICENSE:                MIT

        CONTACT:
          Author                Aristide Grange
          Address               Université de Lorraine
                                Laboratoire LCOMS - UFR MIM
                                3 rue Augustin Fresnel
                                57070 METZ Technopôle
                                France
          Mail                  <author.full.name>@univ-lorraine.fr
    """)) # fmt: skip

    transformations = Transformations(default_params["language"])

    mocodo_group.add_argument("--help",
        action="help",
        help=_("show this help message, then exit"),
    )
    mocodo_group.add_argument("--version",
        action="version",
        version="%(prog)s " + version,
        help=_("display the version number, then exit"),
    )
    mocodo_group.add_argument("--restore",
        action="store_true",
        help=_("recreate a pristine version of the files 'sandbox.mcd' and 'params.json' in the input directory, then exit"),
    )

    io_group.add_argument("--lib",
        metavar="URL",
        nargs="?",
        help=_("remote directory to use as fallback when the input file is not found locally"),
    )
    io_group.add_argument("--output_dir",
        metavar="PATH",
        help=_("the directory of the output files"),
    )
    io_group.add_argument("--encodings",
        metavar="STR",
        nargs="*",
        help=_("one or several encodings to be tried successively when reading the input file"),
    )
    io_group.add_argument("--svg_to",
        choices=["png", "pdf"],
        nargs="+",
        default=[],
        help=_("generate a PNG or a PDF version of the SVG output (requires CairoSVG)"),
    )
    io_group.add_argument("--print_params",
        action="store_true",
        help=_("display the contents of the parameter file, then exit"),
    )
    io_group.add_argument("--reuse_geo",
        action="store_true",
        help=_("reuse the geometry file of the previous execution"),
    )
    io_group.add_argument("--uid_suffix",
        metavar="INT",
        type=non_negative_integer,
        default=0,
        help=_("discriminate between multiple SVG of the same interactive diagram"),
    )
    io_group.add_argument("--select",
        choices=["all", "mcd", "rw", "source", "text", "code", "mocodo", "cv", "mld", "ddl", "sql"],
        nargs="*",
        default=argparse.SUPPRESS, # causes no attribute to be added if the argument was not present
        help=_("under Jupyter Notebook, explicitely state the categories of results to display"),
    )
    io_group.add_argument("--defer",
        metavar="STR",
        nargs="*",
        default=argparse.SUPPRESS,
        help=_("use an external web-service to further convert the conversion results into the given graphical formats"),
    )
    io_group.add_argument("--mld",
        action="store_true",
        help=_("backward compatibility alias for '-t' (with no arguments). Same as '-t markdown' but, under Jupyter Notebook, does not prevent the rendering of the conceptual diagram in the cell output"),
    )
    io_group.add_argument("--is_magic",
        action="store_true",
        help=argparse.SUPPRESS, # don't show this argument in the help message
    )
    io_group.add_argument("--transform", "-t",
        metavar="STR",
        nargs="*",
        type=transformations.extract_transformation_subargs,
        default=argparse.SUPPRESS,
        help=_("make a new version of the MCD by applying sequentially the given rewriting operations, and/or convert it into the given formats or languages. Under Jupyter Notebook, '-T' respectively replaces the current cell by the textual result, or copies it into the clipboard (pip3 install pyperclip)"),
    )
    io_group.add_argument("--Transform", "--TRANSFORM", "-T",
        metavar="STR",
        nargs="*",
        type=transformations.extract_transformation_subargs,
        help=argparse.SUPPRESS, # don't show this argument in the help message
    )
    io_group.add_argument("--seed",
        metavar="FLOAT",
        nargs="?",
        type=float,
        help=_("initial value for the random number generator"),
    )
    io_group.add_argument("--title",
        metavar="STR",
        default=_("Untitled"),
        type=str,
        help=_("name of the model, used at various places (file system, database, etc.)"),
    )

    aspect_group.add_argument("--df",
        metavar="STR",
        type=str,
        default="DF",
        help=_("the acronym to be circled in a functional dependency"),
    )
    aspect_group.add_argument("--card_format",
        metavar="STR",
        type=str,
        nargs="?",
        default="{min_card},{max_card}",
        help=_("format string for minimal and maximal cardinalities"),
    )
    aspect_group.add_argument("--fk_format",
        metavar="STR",
        type=str,
        nargs="?",
        default="#{label}",
        help=_("format string for foreign keys in relational diagram"),
    )
    aspect_group.add_argument("--strengthen_card",
        metavar="STR",
        type=str,
        nargs="?",
        default="_1,1_",
        help=_("string for relative cardinalities"),
    )
    aspect_group.add_argument("--flex",
        metavar="FLOAT",
        type=float,
        default=0.75,
        help=_("flex straight legs whose cardinalities may collide"),
    )
    aspect_group.add_argument("--colors",
        metavar="STEM_OR_PATH",
        default="bw",
        help=_("the color palette to use when generating the drawing. Name (without extension) of a file located in the directory 'colors', or path to a personal file"),
    )
    aspect_group.add_argument("--shapes",
        metavar="STEM_OR_PATH",
        help=_("specification of the fonts, dimensions, etc. Name (without extension) of a file located in the directory 'shapes', or path to a personal file"),
    )
    aspect_group.add_argument("--scale",
        metavar="RATE",
        type=scale,
        default=1,
        help=_("scale the diagram by the given factor"),
    )
    aspect_group.add_argument("--adjust_width",
        metavar="RATE",
        type=scale,
        default=1,
        help=_("scale all calculated text widths by the given factor"),
    )
    aspect_group.add_argument("--detect_overlaps",
        action="store_true",
        help=_("raise an error when horizontal or vertical legs overlap"),
    )
    aspect_group.add_argument("--no_assoc_ids",
        action="store_true",
        help=_("forbid the use of identifiers in associations (according to the Merise standard)"),
    )
    aspect_group.add_argument("--gutters",
        metavar="STR",
        nargs="+",
        type=extract_subargs,
        default=argparse.SUPPRESS, 
        help=_("set the visibility and the contents of the lateral gutters"),
    )

    parser.set_defaults(**default_params)
    params = vars(parser.parse_args(remaining_args))
    params["script_directory"] = SCRIPT_DIRECTORY
    params["output_name"] = Path(params["output_dir"]) / Path(params["input"]).stem
    params["rewrite"] = transformations.operations["rw"]
    params["convert"] = transformations.operations["cv"]
    params["args_to_delete"] = transformations.args_to_delete
    params["opt_to_restore"] = transformations.opt_to_restore
    params["op_tk_rewritings"] = transformations.op_tk_rewritings
    params["redirect_output"] = ("-T" in remaining_args or "--Transform" in remaining_args)
    params["keys_to_hide"] = [
        "keys_to_hide",
        "script_directory",
        "output_name",
        "rewrite",
        "convert",
        "args_to_delete",
        "opt_to_restore",
        "op_tk_rewritings",
        "redirect_output",
        "params_path",
        "Transform",
        "is_magic",
        "print_params",
        "reuse_geo",
        "input",
    ]

    lib = "https://mocodo.net/web/lib"
    if params["lib"]:
        parse_url = importlib.import_module("urllib.parse").urlparse
        if parse_url(params["lib"]).scheme:
            lib = params["lib"]
    params["lib"] = lib

    if params["seed"] is not None:
        random.seed(params["seed"])
    
    with contextlib.suppress(UnicodeDecodeError, AttributeError):
        params["title"] = params["title"].decode("utf8")
    
    return params



__import__("sys").path[0:0] = ["."]

import contextlib
from ..parse_mcd import Token, Visitor
from ..tools.parser_tools import parse_source, reconstruct_source, first_child

class Exploder(Visitor):

    def __init__(self, subargs, params):
        arity = 3
        with contextlib.suppress(ValueError, KeyError):
            arity = float(subargs["arity"])
        self.empty_only = (arity == 2.5)
        self.threshold = int(arity)
        self.allow_weak = "weak" in subargs
        if self.allow_weak:
            # Don't create an identifier for the new weak entity
            self.explosion_template = ":"
            # Prefix the potential first attribute with an underscore
            self.explosion_prefix = " _"
            # Strenghten the new legs
            self.explosion_card = "_11"
        else:
            # Create a strong identifier for the new entity
            self.explosion_template = ": id. {}"
            # Separate the potential first attribute with a comma
            self.explosion_prefix = ", "
            # Don't strengthen the new legs
            self.explosion_card = "11"
        self.df_label = params["df"]

    def line(self, tree):
        # It is not possible to use a method `assoc_clause` since the amount of indentation needs
        # to be known for the added clauses.

        # Guard: store the indentation and abort if the clause is not an association.
        indent = next(tree.find_data("indent"), None)
        indent = indent.children[0].value if indent else ""
        tree = next(tree.find_data("assoc_clause"), None)
        if tree is None:
            return

        # Guard: ensure that there are enough legs and all their max cards are N.
        cards = [node.children[0].value for node in tree.find_data("card")]
        if len(cards) < self.threshold or any(card[1] != "N" for card in cards):
            return
        
        # Guard: on demand, avoid processing binary associations with no attributes.
        if self.empty_only and not first_child(tree, "typed_attr"):
            return

        # Guard: don't explode a clustered association if weak is not allowed
        card_prefixes = [node.children[0].value == "/" for node in tree.find_data("card_prefix")]
        if any(card_prefixes) and not self.allow_weak:
            return
        
        # Guard: don't explode an association of several clusters
        if sum(card_prefixes) > 1:
            return
        
        # Back the legs up as a list of strings.
        legs = list(map(reconstruct_source, tree.find_data("assoc_leg")))

        # Convert the original association into an entity.
        assoc_name = first_child(tree, "assoc_name_def").children[0]
        state = "Waiting for first comma"
        children = []
        for child in tree.children:
            if isinstance(child, Token):
                if child.type == "COMMA" and state == "Waiting for first comma":
                    child.value = self.explosion_template.format(assoc_name.lower())
                    children.append(child)
                    state = "Scanning legs"
                    continue
                if child.type == "COLON": # There were at least one attribute
                    state = "Scanning attributes"
                    child.value = self.explosion_prefix
                if child.type == "NL": # useful where there are no attributes
                    state = "EOL"
            if state != "Scanning legs":
                children.append(child)
        tree.children = children

        # Construct the clauses of the new associations
        clauses = []
        prefix = first_child(tree, "box_def_prefix")
        for leg in legs:
            card = self.explosion_card
            # TODO: check this logic
            if leg.startswith("/"): # when there is a cluster
                leg = leg[1:] # suppress it
                card = "11" # and switch to a non-strenghtening leg
            clauses.append(f"{indent}{prefix}{self.df_label}, {card} {assoc_name}, {leg}")

        # Suffix the resulting string to the last node (a "NL" token")
        tree.children[-1].value += "\n".join(clauses) + "\n"


def run(source, subargs=None, params=None, **kargs):
    subargs = subargs or {}
    params = params or {"df": "DF"}
    tree = parse_source(source)
    visitor = Exploder(subargs, params)
    visitor.visit(tree)
    return reconstruct_source(tree)



import json
from pathlib import Path


def font_metrics_factory(params):
    class FontMetrics:

        data = json.loads(Path(params["script_directory"]).joinpath("resources").joinpath("font_metrics.json").read_text(encoding="utf8"))

        def __init__(self, font):
            if font["family"] not in self.data["fonts"]:
                font["family"] = "Courier New"
            metrics = self.data["fonts"][font["family"]]
            self.font_height = int(round(metrics["height"] * font["size"] / self.data["size"]))
            alphabet = self.data["alphabet"]
            self.width = {c: ord(x) for (c, x) in zip(alphabet, metrics.get("widths", []))}
            self.default_width = metrics["default"]
            self.ratio = font["size"] * metrics.get("correction", 1) / self.data["size"]
            self.ratio *= params["adjust_width"]

        def get_pixel_height(self):
            return self.font_height

        def get_pixel_width(self, s):
            width = sum(self.width.get(c, self.default_width) for c in s)
            return int(round(self.ratio * width)) + 1

    return FontMetrics



import argparse
import contextlib
import importlib
import json
import re
import shlex
import warnings
from base64 import b64encode
from itertools import takewhile
from pathlib import Path

from IPython import get_ipython
from IPython.display import HTML, SVG, Code, Image, Markdown, display

from .__main__ import Printer, Runner
from .mocodo_error import MocodoError

# ANSI color codes
OK = "\033[92m"
WARNING = "\033[1m\033[38;5;166m"
FAIL = "\033[1m\033[91m"
RESET = "\033[0m"

IPYTHON = get_ipython()

def update_cell(content):
    IPYTHON.set_next_input(content, replace=True)

PARAM_TEMPLATE = '''\
# You may edit and run the following lines
import json, pathlib
params = """\\
{stdoutdata}"""
try:
    json.loads(params)
except:
    raise RuntimeError("Invalid JSON. Check your syntax on https://jsonlint.com.")
pathlib.Path("{output_dir}/params.json").write_text(params, encoding="utf8");'''

OUTPUT_PART_HEADER = re.sub("  +", "", """
<div style="position: relative; height: 3ex; background-color: transparent">
  <hr style="margin: 1ex 0 0 0; border-top: 1px solid #BBB">
  <span style="position: absolute; right: 0; color: #BBB">
    <tt>
        {label}
    </tt>
  </span>
</div>
""")

def read_and_cleanup_text(path):
    text = path.read_text(encoding="utf8")
    text = re.sub(r"(?m).+Generated by Mocodo.+\n+", "", text)
    return text

def display_converted_file(path, hide_header):
    if not (hide_header and path.name.endswith("_mld.md")):
        display(Markdown(OUTPUT_PART_HEADER.format(label=path.relative_to(Path.cwd()))))
    extension=path.suffix[1:]
    if extension == "svg":
        # Fix a maximum width for SVG images:
        # https://stackoverflow.com/questions/51452569/how-to-resize-rescale-a-svg-graphic-in-an-ipython-jupyter-notebook
        svg = b64encode(path.read_bytes()).decode("utf8")
        # Use Markdown instead of HTML to avoid a grey background.
        display(Markdown(f'<img max-width="100%" src="data:image/svg+xml;base64,{svg}">'))
    elif extension == "md":
        display(Markdown(read_and_cleanup_text(path)))
    elif extension == "png":
        display(Image(filename=path, unconfined=False))
    elif extension == "html":
        text = path.read_text(encoding="utf8")
        text = re.sub('<!-- TO_BE_DELETED_BY_MOCODO_MAGIC -->.+\n', "", text)
        display(HTML(text))
    elif extension == "tex":
        display(Code(filename=path, language="latex"))
    elif extension == "url":
        print(path.read_text(encoding="utf8")) # make the link clickable
    elif extension == "tsv":
        try:
            df = importlib.import_module("pandas").read_csv(path, sep="\t", header=0, index_col=False)
            df = df.style.set_properties(**{'text-align': 'left'}).set_table_styles([ dict(selector='th', props=[('text-align', 'left')] ) ])
            display(df)
        except ImportError:
            display(Markdown(f'```\n{path.read_text(encoding="utf8")}\n```'))
    else:
        display(Markdown(f"```{extension}\n{read_and_cleanup_text(path)}\n```"))

def mocodo(line, cell=""):
    """
    Mocodo IPython magic extension

    Magic methods:
        %mocodo [command line options]
        %%mocodo [command line options]
        < MCD ... >

    Usage:
        %load_ext mocodo
    """

    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument("--input", "-i")
    parser.add_argument("--output_dir")
    (args, remaining_args) = parser.parse_known_args(shlex.split(line))
    remaining_args = list(takewhile(lambda x: not x.startswith("#"), remaining_args))
    new_args = remaining_args[:]

    mocodo_notebook_dir = Path.cwd()
    if mocodo_notebook_dir.name != "mocodo_notebook":
        mocodo_notebook_dir = mocodo_notebook_dir / "mocodo_notebook"
        mocodo_notebook_dir.mkdir(parents=True, exist_ok=True)
    if not args.input:
        input_path = mocodo_notebook_dir / "sandbox.mcd"
        input_path.write_text(cell, encoding="utf8")
    else:
        input_path = Path(args.input)
        if not input_path.suffix:
            input_path = input_path.with_suffix(".mcd")

    if not args.output_dir:
        output_dir = mocodo_notebook_dir
    else:
        output_dir = Path(args.output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)
    output_path_radical = output_dir / input_path.stem

    remaining_args.extend([
        "--input", str(input_path),
        "--output_dir", str(output_dir),
        "--is_magic",
    ]) # may override user's provided options

    stdoutdata = ""
    stderrdata = ""
    printer = Printer(accumulate=True)
    run = Runner(remaining_args, printer)
    try:
        stdoutdata = run()
    except MocodoError as err:
        stderrdata = str(err)
    except SystemExit: # raised by argparse when --help is used
        return
    
    if "--print_params" in remaining_args:
        update_cell(PARAM_TEMPLATE.format(stdoutdata=stdoutdata, output_dir=output_dir.relative_to(Path.cwd())))
        return
    
    response_path = Path(f"{output_path_radical}_response_for_magic_command.json")
    try: 
        response = json.loads(response_path.read_text(encoding="utf8"))
    except (json.decoder.JSONDecodeError, FileNotFoundError):
        response = {}
    finally:
        with contextlib.suppress(FileNotFoundError): # From Python 3.8, use the missing_ok argument
            response_path.unlink()
    
    rewritten_source = response.get("rewritten_source", "")
    redirect_output = response.get("redirect_output", False)
    converted_file_paths = response.get("converted_file_paths", [])

    if stderrdata:
        message = stderrdata
        if rewritten_source:
            message = f"{rewritten_source}\n\n{message}"
        warnings.formatwarning = lambda x, *args, **kargs : str(x)
        warnings.warn(message)
        return
    
    if rewritten_source and not rewritten_source.startswith("%%mocodo"):
        new_args = " ".join(filter(lambda x: x not in response["args_to_delete"], new_args))
        new_args = response["opt_to_restore"] + new_args
        rewritten_source = f"%%mocodo{new_args}\n{rewritten_source}"
    
    for select in response.get("select", []):
        if select == "mcd":
            svg_path = output_path_radical.with_suffix(".svg")
            if svg_path.is_file() and input_path.stat().st_mtime <= svg_path.stat().st_mtime:
                display(SVG(filename=svg_path))
        elif select == "rw":
            path = output_path_radical.with_suffix(".mcd")
            display(Markdown(OUTPUT_PART_HEADER.format(label=path.relative_to(output_dir))))
            print(rewritten_source or cell)
        elif select == "cv":
            for converted_file_path in converted_file_paths:
                display_converted_file(Path(converted_file_path), hide_header=response.get("mld"))

    if redirect_output:
        if rewritten_source:
            update_cell(rewritten_source)
        if converted_file_paths:
            # Copy the last converted file source to the clipboard
            with contextlib.suppress(ImportError):
                pyperclip = importlib.import_module("pyperclip")
                converted_file_path = Path(converted_file_paths[-1])
                pyperclip.copy(converted_file_path.read_text(encoding="utf8"))
                print(f'⧉ {OK}The contents of "{converted_file_path.relative_to(output_dir)}" has been copied to the clipboard.{RESET}')



from ..tools.parser_tools import parse_source, Token

def ast_to_yaml(node, indent=""):
    indent += "  "
    if isinstance(node, Token):
        yield f"{indent}- type: {node.type}"
        yield f"{indent}  value: {repr(node.value)}"
    else:
        yield f"{indent}- type: {node.data}"
        yield f"{indent}  children:"
        for child in node.children:
            yield from ast_to_yaml(child, indent)


def run(source, subargs=None, common=None):
    tree = parse_source(source)
    return {
        "stem_suffix": "_ast",
        "text": "\n".join(ast_to_yaml(tree)),
        "extension": "yaml",
        "to_defer": False,
        "highlight": "yaml",
    }



from collections import defaultdict
from .attribute import *
from .tools.string_tools import rstrip_digit_or_underline, raw_to_bid

class Entity:
    def __init__(self, clause):
        self.source = clause["source"]
        self.raw_name = clause["name"]
        # A protected entity results in a table, even if all its columns are part of its primary key.
        self.is_protected = (clause.get("box_def_prefix") == "+")
        self.bid = raw_to_bid(self.raw_name)
        self.name_view = rstrip_digit_or_underline(self.raw_name)
        self.attributes = clause.get("attrs", [])
        self.legs = []  # iterating over box's legs does nothing if it is not an association
        self.kind = "entity"
        if clause.get("box_def_prefix") == "-":
            self.calculate_size = self.calculate_size_when_invisible
            self.description = lambda *ignored: []
            self.is_invisible = True
        else:
            self.calculate_size = self.calculate_size_when_visible
            self.description = self.description_when_visible
            self.is_invisible = False
        self.has_alt_identifier = False

    def add_attributes(self, legs_to_strengthen, is_child, fk_format):
        weak_entity = bool(legs_to_strengthen)
        self.strengthening_legs = legs_to_strengthen
        for (i, a) in enumerate(self.attributes):
            id_mark = a.get("id_mark","")
            explicit_underscore = "0" in a.get("id_groups", "") or a.get("id_groups", "") == ""
            if a.get("attribute_label", "") == "":
                attribute = PhantomAttribute(a)
            elif is_child:
                attribute = SimpleEntityAttribute(a)
            elif i == 0 and id_mark != "_":
                attribute = WeakAttribute(a) if  weak_entity else StrongAttribute(a)
            elif i == 0 and id_mark == "_" and not explicit_underscore:
                attribute = WeakAttribute(a) if  weak_entity else StrongAttribute(a)
            elif i != 0 and id_mark == "_" and explicit_underscore:
                attribute = WeakAttribute(a) if  weak_entity else StrongAttribute(a)
            else:
                attribute = SimpleEntityAttribute(a)
            attribute.register_foreign_key_status(a, fk_format)
            self.attributes[i] = attribute
        self.candidates = defaultdict(set)
        for a in self.attributes:
            for id_group in a.id_groups:
                self.candidates[id_group].add(a.label)
        self.candidates = dict(self.candidates)
        if len(self.candidates) > 1:
            self.has_alt_identifier = True
        
    def register_boxes(self, boxes):
        self.boxes = boxes
    
    def set_id_gutter_visibility(self, is_visible):
        self.show_id_gutter = is_visible
    
    def calculate_size_when_invisible(self, *ignored):
        self.w = 0
        self.h = 0
    
    def calculate_size_when_visible(self, style, get_font_metrics):
        cartouche_font = get_font_metrics(style["entity_cartouche_font"])
        self.get_cartouche_string_width = cartouche_font.get_pixel_width
        self.cartouche_height = cartouche_font.get_pixel_height()
        attribute_font = get_font_metrics(style["entity_attribute_font"])
        self.attribute_height = attribute_font.get_pixel_height()
        for attribute in self.attributes:
            attribute.calculate_size(style, get_font_metrics)
        cartouche_and_attribute_widths = []
        cartouche_and_attribute_widths.append(self.get_cartouche_string_width(self.name_view))
        cartouche_and_attribute_widths.extend(a.w for a in self.attributes)
        self.id_gutter_width = 0
        if self.show_id_gutter:
            self.id_gutter_width = 2 * style["rect_margin_width"]
            if self.attributes:
                self.id_gutter_width += max(attribute.id_width for attribute in self.attributes)
        self.w = 2 * style["rect_margin_width"] + self.id_gutter_width + max(cartouche_and_attribute_widths)
        self.h = (
            len(self.attributes) * (self.attribute_height + style["line_skip_height"])
            - style["line_skip_height"]
            + 4 * style["rect_margin_height"]
            + self.cartouche_height
        )
        self.w += self.w % 2
        self.h += self.h % 2
        for attribute in self.attributes:
            attribute.set_id_gutter_width(self.id_gutter_width)

    def register_center(self, geo):
        self.cx = geo["cx"][self.bid]
        self.cy = geo["cy"][self.bid]
        self.l = self.cx - self.w // 2
        self.r = self.cx + self.w // 2
        self.t = self.cy - self.h // 2
        self.b = self.cy + self.h // 2

    def description_when_visible(self, style, geo):
        result = []
        result.append(("comment", {"text": f"Entity {self.bid}"}))
        result.append(
            (
                "begin_component",
                {
                    "page": self.page,
                    "visibility": "hidden" if self.page else "visible",
                }
            )
        )
        result.append(("begin_group", {}))
        result.append(
            (
                "rect",
                { # upper part background
                    "x": self.l,
                    "y": self.t,
                    "w": self.w,
                    "h": self.cartouche_height + 2 * style["rect_margin_height"],
                    "color": style["entity_cartouche_color"],
                    "stroke_color": "none",
                    "stroke_depth": 0,
                    "opacity": 1,
                },
            )
        )
        result.append(
            ( # lower part background (with or without a margin for the left gutter)
                "rect",
                { 
                    "x": self.l + self.id_gutter_width,
                    "y": self.t + self.cartouche_height + 2 * style["rect_margin_height"],
                    "w": self.w - self.id_gutter_width,
                    "h": self.h - self.cartouche_height - 2 * style["rect_margin_height"],
                    "color": style["entity_color"],
                    "stroke_color": "none",
                    "stroke_depth": 0,
                    "opacity": 1,
                },
            )
        )
        if self.show_id_gutter:
            result.append(
                ( # id_gutter background
                    "rect",
                    {
                        "x": self.l,
                        "y": self.t + self.cartouche_height + 2 * style["rect_margin_height"],
                        "w": self.id_gutter_width,
                        "h": self.h - self.cartouche_height - 2 * style["rect_margin_height"],
                        "color": style["id_gutter_color"],
                        "stroke_color": "none",
                        "stroke_depth": 0,
                        "opacity": 1,
                    },
                )
            )
            result.append(
                ( # line at the right of the left gutter
                    "line",
                    {
                        "x0": self.l + self.id_gutter_width,
                        "y0": self.t + self.cartouche_height + 2 * style["rect_margin_height"],
                        "x1": self.l + self.id_gutter_width,
                        "y1": self.b,
                        "stroke_color": style["entity_stroke_color"],
                        "stroke_depth": style["inner_stroke_depth"] / 4,
                    },
                )
            )
        result.append(
            ( # outer frame
                "rect",
                {
                    "x": self.l,
                    "y": self.t,
                    "w": self.w,
                    "h": self.h,
                    "color": style["transparent_color"],
                    "stroke_color": style["entity_stroke_color"],
                    "stroke_depth": style["box_stroke_depth"],
                    "opacity": 1,
                },
            )
        )
        result.append(
            ( # line between upper and lower part
                "line",
                {
                    "x0": self.l,
                    "y0": self.t + self.cartouche_height + 2 * style["rect_margin_height"],
                    "x1": self.r,
                    "y1": self.t + self.cartouche_height + 2 * style["rect_margin_height"],
                    "stroke_color": style["entity_stroke_color"],
                    "stroke_depth": style["inner_stroke_depth"],
                },
            )
        )
        result.append(("end", {}))
        result.append(
            ( # cartouche text
                "text",
                {
                    "x": self.cx - self.get_cartouche_string_width(self.name_view) // 2,
                    "y": self.t + style["rect_margin_height"] + style["cartouche_text_height_ratio"] * self.cartouche_height,
                    "text_color": style["entity_cartouche_text_color"],
                    "family": style["entity_cartouche_font"]["family"],
                    "size": style["entity_cartouche_font"]["size"],
                    "text": self.name_view,
                },
            )
        )
        x = self.cx - self.w // 2 + style["rect_margin_width"]
        dx = self.id_gutter_width
        dy = self.cartouche_height + 3 * style["rect_margin_height"] - self.h // 2
        for attribute in self.attributes:
            result.extend(attribute.description(style, x, self.cy, dx, dy))
            dy += self.attribute_height + style["line_skip_height"]
        result.append(("end", {}))
        return result



from math import sqrt

from .attribute import *
from .leg import *
from .tools.string_tools import rstrip_digit_or_underline, raw_to_bid

TRIANGLE_ALTITUDE = sqrt(3) / 2
INCIRCLE_RADIUS = 1 / sqrt(12)


class Inheritance:

    counter = 0

    @classmethod
    def reset_counter(cls):
        cls.counter = 0

    def __init__(self, clause, **params):
        self.source = clause["source"]
        self.raw_name = clause["name"] if clause["name"] != "TX" else "XT"
        Inheritance.counter += 1
        parent_bid = raw_to_bid(clause["legs"][0]["entity"])
        self.bid = f'{parent_bid}_PARENT_#{Inheritance.counter}'
        self.name_view = rstrip_digit_or_underline(clause["name"])
        self.attributes = [InheritanceAttribute(attr) for attr in clause.get("attrs", [])]
        for leg_clause in clause["legs"]:
            leg_clause["kind"] = "-"
            leg_clause["arrow"] = False
        inheritance_arrow = clause["inheritance_arrow"]
        if inheritance_arrow == "<-":
            clause["legs"][0]["arrow"] = True
        elif inheritance_arrow == "->":
            for leg_clause in clause["legs"][1:]:
                leg_clause["arrow"] = True
        elif inheritance_arrow == "<=":
            clause["legs"][0]["arrow"] = True
            for leg_clause in clause["legs"][1:]:
                leg_clause["kind"] = "="
        elif inheritance_arrow == "=>":
            clause["legs"][0]["kind"] = "="
            for leg_clause in clause["legs"][1:]:
                leg_clause["arrow"] = True
        else:
            clause["legs"][0]["arrow"] = True
        self.legs = [InheritanceLeg(self, leg, **params) for leg in clause["legs"]]
        self.kind = inheritance_arrow.replace("--", "-").replace("==", "=")

    def register_boxes(self, boxes):
        self.boxes = boxes

    def calculate_size(self, style, get_font_metrics):
        cartouche_font = get_font_metrics(style["association_cartouche_font"])
        self.get_cartouche_string_width = cartouche_font.get_pixel_width
        self.cartouche_height = cartouche_font.get_pixel_height()
        attribute_font = get_font_metrics(style["association_attribute_font"])
        self.attribute_height = attribute_font.get_pixel_height()
        self.w = self.h = 2 * (style["round_rect_margin_width"] + self.cartouche_height)
        self.w += self.w % 2
        self.h += self.h % 2
        for leg in self.legs:
            leg.calculate_size(style, get_font_metrics)

    def register_center(self, geo):
        self.cx = geo["cx"][self.bid]
        self.cy = geo["cy"][self.bid]
        self.l = self.cx - self.w // 2
        self.r = self.cx + self.w // 2
        self.t = self.cy - self.h // 2
        self.b = self.cy + self.h // 2

    def description(self, style, geo):
        result = []
        result.append(("comment", {"text": f"Inheritance {self.bid}"}))
        result.append(
            (
                "begin_component",
                {
                    "page": self.page,
                    "visibility": "hidden" if self.page else "visible",
                }
            )
        )
        result.extend(self.leg_descriptions(style, geo))
        result.append(("begin_group", {}))
        result.extend(
            [
                (
                    "triangle",
                    {
                        "stroke_depth": style["box_stroke_depth"],
                        "stroke_color": style['association_stroke_color'],
                        "color": style['association_cartouche_color'],
                        "x1": self.cx,
                        "x2": self.l,
                        "x3": self.r,
                        "y1": self.cy - (TRIANGLE_ALTITUDE - INCIRCLE_RADIUS) * self.w,
                        "y2": self.cy + INCIRCLE_RADIUS * self.w,
                        "y3": self.cy + INCIRCLE_RADIUS * self.w,
                    },
                ),
                (
                    "text",
                    {
                        "text": self.name_view,
                        "text_color": style['association_cartouche_text_color'],
                        "x": self.cx - self.get_cartouche_string_width(self.name_view) // 2,
                        "y": self.cy + self.cartouche_height // 3,
                        "family": style["association_cartouche_font"]["family"],
                        "size": style["association_cartouche_font"]["size"],
                    },
                ),
            ]
        )
        result.append(("end", {}))
        result.append(("end", {}))
        return result

    def leg_descriptions(self, style, geo):
        result = []
        for leg in self.legs:
            result.extend(leg.description(style, geo))
        return result



import re

def create_name_to_index():
    """
    Return a function that maps names to unique numbers.
    Used to generate short "names" for Graphviz nodes.
    """
    cache = {}
    def name_to_index(name):
        if name not in cache:
            cache[name] = len(cache) + 1
        return cache[name]
    return name_to_index


def minify_graphviz(text):
    # Suppress comments
    text = re.sub(r"(?m)^ *//.*$", "", text)
    # Suppress empty lines
    text = re.sub(r"(?m)^\s*\n", "", text)
    # Suppress leading spaces
    text = re.sub(r"(?m)^\s+", "", text)
    # Suppress newline before delimiters
    text = re.sub(r"\n([]<>])", r"\1", text)
    # Factorize edges
    while True:
        (text, n) = re.subn(r"(?m)(\d+) (-[->]) ([^[\n]+)\n\1 -- ", r"\1 \2 \3,", text)
        if n == 0:
            break
    # Suppress spaces around arrows
    text = re.sub(r"(?m)(\d+) (-[->]) ", r"\1\2", text)
    # Suppress spaces before opening brackets
    text = re.sub(r"(?m)^(edge|node) *\[", r"\1[", text)
    # Suppress newlines after opening brackets
    text = re.sub(r"\[\n", r"[", text)
    # Supress spaces after \d+ -> \d+
    text = re.sub(r"(?m)^([\d>-]+) +", r"\1", text)
    return text


NODE_OPTIONS_TEMPLATE = """
      shape=none
      fontcolor="{cell_font_color}"
      fontsize={cell_font_size}
      fontname="Helvetica"
""" # This constant is public

TABLE_TEMPLATE = """<table
        border="0"
        cellborder="1"
        cellspacing="0"
        bgcolor="{cell_bg_color}"
        color="{stroke_color}"
        cellpadding="4"
    >
        TABLE_CONTENT_PLACEHOLDER
    </table>
"""

HEADER_TEMPLATE = """<tr>
            <td bgcolor="{header_bg_color}" colspan="COL_COUNT_PLACEHOLDER">
                <font
                    color="{header_font_color}"
                    point-size="{header_font_size}"
                >HEADER_TEXT_PLACEHOLDER</font>
            </td>
        </tr>
"""

SPECIAL_FONT_TEMPLATE = """<font face="Courier">TEXT_PLACEHOLDER</font>"""

ROW_TEMPLATE = """        <tr>{cells}</tr>"""


def row_format_to_td_templates(
    row_format, # a LaTeX-like table format string (e.g. "|l|c|r|")
    ALIGN={
        "l": ' align="left"',
        "c": "",
        "r": ' align="right"',
    },
):
    """
    Return a list of td element templates from a LaTeX-table-like format string.
    NB: an uppercase letter means "use the special font".
    """
    acc = []
    row_format = f" {row_format} "
    for (i, symbol) in enumerate(row_format[1:-1], 1):
        text = "TEXT_PLACEHOLDER"
        if symbol.isupper():
            text = SPECIAL_FONT_TEMPLATE
            symbol = symbol.lower()
        if symbol in ALIGN:
            align = ALIGN[symbol]
            sides = ' sides="TB'
            if row_format[i - 1] == "|":
                sides += "L"
            if row_format[i + 1] == "|":
                sides += "R"
            sides += '"'
            if sides == ' sides="TBLR"':
                sides = ""
            acc.append(f"<td{align}{sides}>{text}</td>")
    return acc

def table_as_label(header_text, rows, row_format, style):
    """Return a table has an HTML-like Graphviz label."""
    header = HEADER_TEMPLATE.format(**style)
    header = header.replace("HEADER_TEXT_PLACEHOLDER", header_text)
    td_templates = row_format_to_td_templates(row_format)
    header = header.replace("COL_COUNT_PLACEHOLDER", str(len(td_templates)))
    trs = []
    for row in rows:
        cells = []
        for (text, td_template) in zip(row, td_templates):
            cells.append(td_template.replace("TEXT_PLACEHOLDER", text))
        trs.append(ROW_TEMPLATE.format(cells="".join(cells)))
    trs = "\n".join(trs)
    table = TABLE_TEMPLATE.format(**style)
    table = table.replace("TABLE_CONTENT_PLACEHOLDER", header + trs)
    return table

if __name__ == "__main__":
    header_text = "CLIENT"
    rows = [
        ["PK", "id. client", "varchar(8)"],
        ["", "nom", "varchar(30)"],
        ["", "prénom", "varchar(30)"],
        ["", "âge", "int"],
    ]
    row_format = "|cl|L|"
    style = {
        "cell_bg_color": "#c0d4ff",
        "stroke_color": "#578dff",
        "header_bg_color": "#97b8ff",
        "header_font_color": "#131114",
        "header_font_size": 14,
        "cell_font_color": "#3e3c42",
        "cell_font_size": 12,
    }
    node_options = NODE_OPTIONS_TEMPLATE.format(**style)
    table = table_as_label(header_text, rows, row_format, style)
    print("digraph {")
    print(f"node [{node_options}]")
    print(f'1 [label=<{table}>]')
    print("}")



from pathlib import Path
import re

from .mocodo_error import MocodoError
from .version_number import version

try:
    from cairosvg import svg2png, svg2pdf
except:
    def svg2png(**kargs):
        raise MocodoError(13, _("PNG and PDF generation requires cairosvg to be installed")) # fmt: skip
    svg2pdf = svg2png


def main(mcd, common, suppress_mocodo_header=False):
    mocodo_header = "" if suppress_mocodo_header else f"<!-- Generated by Mocodo {version} -->\n"
    style = common.load_style()
    mcd.calculate_size(style)
    geo = mcd.calculate_or_retrieve_geo(common.params)
    try:
        description = mcd.description(style, geo)
    except KeyError: # retry silently with a fresh geo file if the previous one concerns another MCD
        Path(f"{common.params['output_name']}_geo.json").unlink()
        geo = mcd.calculate_or_retrieve_geo(common.params)
        description = mcd.description(style, geo)
    description = [
        (
            "preamble",
            {
                "width": geo["width"],
                "height": geo["height"],
                "total_height": geo["height"] + (mcd.page_count > 1 and style["note_overlay_height"]),
            }
        ),
        (
            "background",
            {
                "width": geo["width"],
                "height": geo["height"],
                "background_color": style["background_color"],
            }
        ),
        *description,
    ]
    has_note_card = False
    tabs = 0
    categories = { # the order is that of the drawing in the SVG
        "": [],
        "Constraint": [],
        "Inheritance": [],
        "Association": [],
        "Entity": [],
        "Link": [],
        "Notes": [],
        "Pager": [],
    }
    category = ""
    for (key, mapping) in description:
        mapping["mcd_uid"] = mcd.uid
        has_note_card |= key.endswith("with_note")
        if key == "comment":
            category = mapping["text"].partition(" ")[0]
        for (k, v) in mapping.items():
            if isinstance(v, float):
                mapping[k] = int(v) if v.is_integer() else round(v, 2)
            elif isinstance(v, str):
                mapping[k] = html_escape(v)
            elif v is None:
                mapping[k] = "none"
        mapping["mocodo_header"] = mocodo_header
        # print(key, mapping, end="\n")
        tabs -= (key == "end")
        categories[category].append('\t' * tabs + svg_elements[key].format_map(mapping))
        tabs += key.startswith("begin")
    if not has_note_card:
        del categories["Notes"]
    text = "\n".join(sum(categories.values(), [])) + "\n</svg>"
    is_interactive = categories.get("Notes") or categories.get("Pager")
    if not is_interactive:
        text = re.sub(r'<g class="page_0_\w{8}(_\d+)? diagram_page" visibility="visible">', "<g>", text)
    path = Path(f"{common.params['output_name']}.svg")
    path.write_text(text, encoding="utf8")
    resulting_paths = [path]
    if categories.pop("Notes", []) + categories.pop("Pager"): # don't use `or` to avoid short-circuit
        text = "\n".join(sum(categories.values(), [])) + "\n</svg>"
        text = re.sub(
            r"(?m)^<\?xml .+\n<svg .+",
            svg_elements["preamble"].format(mocodo_header=mocodo_header, total_height=geo["height"], **geo),
            text
        )
        text = re.sub(r"(?m)^<g class.+", "<g>", text)
        text = re.sub(r' (onmouseover|onmouseout|style|id)=".+?"', "", text)
        path = Path(f"{common.params['output_name']}_static.svg")
        path.write_text(text, encoding="utf8")
        resulting_paths.append(path)
    svg = bytes(text, "utf-8")
    for (format, function) in (("png", svg2png), ("pdf", svg2pdf)):
        if format in common.params["svg_to"]:
            path = Path(f"{common.params['output_name']}.{format}")
            function(bytestring=svg, write_to=str(path))
            resulting_paths.append(path)
    return resulting_paths



def html_escape(
    text,
    table={
        "&": "&amp;",
        ">": "&gt;",
        "<": "&lt;",
        '"': "&quot;",
        "'": "’",  # neither &#39; nor &apos; make the job in notes
        "-": "&#45;", # prevent double hyphen in SVG comments: https://stackoverflow.com/questions/10842131/xml-comments-and
}):
    return "".join(table.get(c, c) for c in text)

svg_elements = {
    "preamble":         """<?xml version="1.0" encoding="utf-8"?>\n{mocodo_header}<svg width="{width}" height="{total_height}" viewBox="0 0 {width} {total_height}" xmlns="http://www.w3.org/2000/svg">""",
    "background":       """<rect x="0" y="0" width="{width}" height="{height}" fill="{background_color}" stroke="none" stroke-width="0"/>""",
    "comment":          """\n<!-- {text} -->""",
    "begin_component":  """<g class="page_{page}_{mcd_uid} diagram_page" visibility="{visibility}">""",
    "begin_group":      """<g>""",
    "end":              """</g>""",
    "text":             """<text x="{x}" y="{y}" fill="{text_color}" font-family="{family}" font-size="{size}">{text}</text>""",
    "text_above_note":  """<text x="{x}" y="{y}" fill="{text_color}" font-family="{family}" font-size="{size}" style="pointer-events: none;">{text}</text>""",
    "text_with_note":   """<text x="{x}" y="{y}" fill="{text_color}" font-family="{family}" font-size="{size}" onmouseover="show_{mcd_uid}(evt,'{note}')" onmouseout="hide_{mcd_uid}(evt)" style="cursor: pointer;">{text}</text>""",
    "line":             """<line x1="{x0}" y1="{y0}" x2="{x1}" y2="{y1}" stroke="{stroke_color}" stroke-width="{stroke_depth}"/>""",
    "dash_line":        """<line x1="{x0}" y1="{y0}" x2="{x1}" y2="{y1}" stroke="{stroke_color}" stroke-width="{stroke_depth}" stroke-dasharray="{dash_width}"/>""",
    "dot_line":         """<line x1="{x0}" y1="{y0}" x2="{x1}" y2="{y1}" stroke="{stroke_color}" stroke-width="{stroke_depth}"  stroke-dasharray="0,{dash_gap}" stroke-linecap="round"/>""",
    "rect":             """<rect x="{x}" y="{y}" width="{w}" height="{h}" fill="{color}" stroke="{stroke_color}" stroke-width="{stroke_depth}" opacity="{opacity}"/>""",
    "polygon":          """<path d="{path}" fill="{color}" stroke="{stroke_color}" stroke-width="{stroke_depth}" opacity="{opacity}"/>""",
    "dot_polygon":      """<path d="{path}" fill="{color}" stroke="{stroke_color}" stroke-width="{stroke_depth}" stroke-dasharray="0,{dash_gap}" stroke-linecap="round"/>""",
    "circle":           """<circle cx="{cx}" cy="{cy}" r="{r}" stroke="{stroke_color}" stroke-width="{stroke_depth}" fill="{color}"/>""",
    "circle_with_note": """<circle cx="{cx}" cy="{cy}" r="{r}" stroke="{stroke_color}" stroke-width="{stroke_depth}" fill="{color}" onmouseover="show_{mcd_uid}(evt,'{note}')" onmouseout="hide_{mcd_uid}(evt)" style="cursor: pointer;"/>""",
    "pager_dot":        """<circle cx="{cx}" cy="{cy}" r="{r}" fill="{color}" id="pager_dot_{page}_{mcd_uid}" class="pager_dot" stroke-width="0" onclick="switch_page_visibility_{mcd_uid}(evt,{page})" style="cursor: pointer;"/>""",
    "triangle":         """<polygon points="{x1} {y1} {x2} {y2} {x3} {y3}" stroke="{stroke_color}" stroke-width="{stroke_depth}" fill="{color}"/>""",
    "arrow":            """<polygon points="{x0} {y0} {x1} {y1} {x2} {y2} {x3} {y3}" fill="{stroke_color}" stroke-width="0"/>""",
    "curve":            """<path d="M{x0} {y0} C{x1} {y1} {x2} {y2} {x3} {y3}" fill="none" stroke="{stroke_color}" stroke-width="{stroke_depth}"/>""",
    "round_rect":       """<rect x="{x}" y="{y}" width="{w}" height="{h}" fill="{color}" rx="{radius}" stroke="{stroke_color}" stroke-width="{stroke_depth}"/>""",
    "upper_round_rect": """<path d="M{x0} {y0} a{r} {r} 90 0 1 {r} {r} V{y1} h-{w} V{y2} a{r} {r} 90 0 1 {r} -{r}" fill="{color}" stroke="{stroke_color}" stroke-width="{stroke_depth}"/>""",
    "lower_round_rect": """<path d="M{x0} {y0} v{y1} a{r} {r} 90 0 1 -{r} {r} H{x1} a{r} {r} 90 0 1 -{r} -{r} V{y0} H{w}" fill="{color}" stroke="{stroke_color}" stroke-width="{stroke_depth}"/>""",
    "notes":            """<script type="text/ecmascript">
                            <![CDATA[
                            	function show_{mcd_uid}(evt, text) {{
                            		var pos = (evt.target.getAttribute("y") < {height_threshold}) ? "bottom" : "top";
                            		var note = document.getElementById(pos + "_note_{mcd_uid}");
                            		note.textContent = text;
                            		note.setAttributeNS(null, "visibility", "visible");
                            		document.getElementById(pos + "_overlay_{mcd_uid}").setAttributeNS(null, "visibility", "visible");
                            	}}
                            	function hide_{mcd_uid}(evt) {{
                            		document.getElementById("top_note_{mcd_uid}").setAttributeNS(null, "visibility", "hidden");
                            		document.getElementById("top_overlay_{mcd_uid}").setAttributeNS(null, "visibility", "hidden");
                            		document.getElementById("bottom_note_{mcd_uid}").setAttributeNS(null, "visibility", "hidden");
                            		document.getElementById("bottom_overlay_{mcd_uid}").setAttributeNS(null, "visibility", "hidden");
                            	}}
                            ]]>
                            </script>
                            <rect id="top_overlay_{mcd_uid}" x="0" y="0" width="100%" height="{overlay_height}" fill="{color}" stroke-width="0" opacity="{opacity}" visibility="hidden"/>
                            <text id="top_note_{mcd_uid}" text-anchor="middle" x="{x}" y="{y_top}" fill="{text_color}" font-family="{font_family}" font-size="{font_size}" visibility="hidden"></text>
                            <rect id="bottom_overlay_{mcd_uid}" x="0" y="{y}" width="100%" height="{overlay_height}" fill="{color}" stroke-width="0" opacity="{opacity}" visibility="hidden"/>
                            <text id="bottom_note_{mcd_uid}" text-anchor="middle" x="{x}" y="{y_bottom}" fill="{text_color}" font-family="{font_family}" font-size="{font_size}" visibility="hidden"></text>""".replace("    ", ""),
    "pager":            """<script type="text/ecmascript">
                            <![CDATA[
                            	function switch_page_visibility_{mcd_uid}(evt, page) {{
                            		for (var i = 0; i < {page_count}; i++) {{
                            			components = document.getElementsByClassName(`page_${{i}}_{mcd_uid}`);
                            			for (var j = 0; j < components.length; j++) {{
                            				components[j].setAttributeNS(null, "visibility", i <= page ? "visible" : "hidden");
                            			}};
                            			dot = document.getElementById(`pager_dot_${{i}}_{mcd_uid}`);
                            			dot.setAttributeNS(null, "fill", i <= page ? "gray" : "lightgray");
                            		}}
                            	}}
                            ]]>
                            </script>""".replace("    ", ""),
}


if __name__ == "__main__":
    from .argument_parser import parsed_arguments
    from .common import Common
    from .mcd import Mcd

    clauses = """
        CLIENT: Réf. client, Nom, Prénom, Adresse
        PASSER, 0N CLIENT, 11 COMMANDE
        COMMANDE: Num commande, Date, Montant
        INCLURE, 1N COMMANDE, 0N PRODUIT: Quantité
        PRODUIT: Réf. produit, Libellé, Prix unitaire
    """
    params = parsed_arguments()
    common = Common(params)
    mcd = Mcd(clauses, **params)
    main(mcd, common)



from pathlib import Path
import re

from ..parse_mcd import Visitor
from ..tools.parser_tools import first_child, parse_source, reconstruct_source
from ..tools.string_tools import ascii, snake, strip_surrounds

class CreateTypePlaceholder(Visitor):

    def __init__(self, subsubarg):
        self.default = strip_surrounds(subsubarg, "[]")

    def typed_attr(self, tree):
        if len(tree.children) == 1:
            tree.children[0].children[0].value += f" [{self.default}]"

def create_type_placeholders(source, subsubarg):
    visitor = CreateTypePlaceholder(subsubarg)
    tree = parse_source(source)
    visitor.visit(tree)
    return reconstruct_source(tree)

def read_default_datatypes(resource_dir, language="en"):
    path = Path(resource_dir, f"default_datatypes_{language}.tsv")
    if not path.exists():
        return
    for line in path.read_text(encoding="utf8").splitlines():
        fields = line.split("\t")
        if len(fields) < 2:
            continue
        yield tuple(fields[:2])

class GuessType(Visitor):

    def __init__(self,params):
        resource_dir = Path(params["script_directory"], "resources")
        self.field_types = dict(read_default_datatypes(resource_dir)) # English always used as fallback
        language = params.get("language", "en")
        if language != "en":
            self.field_types.update(read_default_datatypes(resource_dir, language))
        # Convert the dict to a list of tuples, and sort them by decreasing length
        # so that the longest match is found first.
        self.field_types = sorted(self.field_types.items(), key=lambda x: -len(x[0]))
        # Precompile the regexes
        self.field_types = [(re.compile(k).search, v) for (k, v) in self.field_types]

    def typed_attr(self, tree):
        n = len(tree.children)
        # Possible lengths of the children list:
        # 0: empty attribute
        # 1: attribute without datatype
        # 3: attribute with empty datatype
        # 4: attribute with non-empty datatype
        if n not in (1, 3):
            return
        attr = first_child(tree, "attr")
        needle = snake(ascii(attr)).replace("_", " ").lower()
        for (found, datatype) in self.field_types:
            if found(needle):
                break
        else:
            datatype = ""
        if n == 1: # concatenate to the attribute label the datatype between brackets
            tree.children[0].children[0].value += f" [{datatype}]"
        else: # insert the datatype between brackets
            tree.children[1].value += datatype


def guess_types(source, params):
    visitor = GuessType(params)
    tree = parse_source(source)
    visitor.visit(tree)
    return reconstruct_source(tree)



from ..tools.string_tools import rstrip_digit_or_underline
from ..tools.parser_tools import parse_source

def run(source, id_prefix):
    if id_prefix is None:
        id_prefix = "id. "
    tree = parse_source(source)
    entity_name_refs =set(node.children[0].children[0].value for node in tree.find_data("entity_name_ref"))
    entity_name_defs = set(node.children[0].children[0].value for node in tree.find_data("entity_name_def"))
    new_entity_names = entity_name_refs - entity_name_defs
    new_entity_clauses = [source]
    for ent in new_entity_names:
        lower_ent = rstrip_digit_or_underline(ent.lower())
        if lower_ent in ("date", "calendrier", "calendar"):
            clause = f"{ent}: date"
        elif lower_ent == "période":
            clause = f"{ent}: début, _fin"
        elif lower_ent == "period":
            clause = f"{ent}: start, _end"
        else:
            clause = f"{ent}: {id_prefix}{lower_ent}, "
        new_entity_clauses.append(clause)
    return "\n".join(new_entity_clauses)



import itertools
import math


class Grid(list):
    
    def __init__(self, max_number_of_nodes=100, min_balance=0.5):
        """Create a list of couples of dimensions.
        
        The nth couple is the tightest balanced rectangle having at least n cells.
        A balanced rectangle satisfies: length / width > min_balance.
        [None, (1, 1), (2, 2), (2, 2), (2, 2), (3, 2), (3, 2), (3, 3), ...]
        """
        list.__init__(self)
        max_square_side = int(math.ceil(math.sqrt(max_number_of_nodes)))
        for n in range(max_square_side * max_square_side, 0, -1):
            for i in range(n, int(math.ceil(math.sqrt(n))) - 1, -1):
                if n % i == 0 and n / i / i > min_balance:
                    self.insert(0, (i, n // i))
                    break
            else:
                self.insert(0, self[0])
        self.insert(0, None)
        self[2:4] = [(2, 1), (3, 1)] # tweak the 2- and 3-box grids
    
    def get_nth_next(self, index, nth):
        """Return the nth next distinct rectangle after the index-th one."""
        acc = set()
        for i in itertools.count(index):
            acc.add(self[i])
            if len(acc) > nth:
                return self[i]
    
    def get_markdown(self):
        """Return the grid as a nicely formatted table. For the docs only."""
        (w, h) = map(max, zip(*self[1:]))
        result = \
            [
                [""] + 
                list(map(lambda i: "**%s**" % i, range(1, w + 1)))
            ] + \
            [["---"] * (w + 1)] + \
            [
                ["**%s**" % row] + 
                [""] * w for row in range(1, h + 1)
            ]
        for (n, (col, row)) in enumerate(self[1:], 1):
            result[row + 1][col] += (", %s" % n if result[row + 1][col] else str(n))
        return "\n".join(["| %s |" % " | ".join(row) for row in result])


if __name__ == "__main__":
    grid = Grid()
    print(grid.get_markdown())



from .mocodo_error import MocodoError


class DiagramLink:
    def __init__(self, entities, foreign_entity, foreign_key):
        self.foreign_entity = foreign_entity
        self.foreign_key = foreign_key
        try:
            self.primary_entity = entities[foreign_key.primary_entity_bid]
        except KeyError:
            raise MocodoError(14, _('Attribute "{attribute}" in entity "{entity_1}" references an unknown entity "{entity_2}".').format(attribute=foreign_key.label, entity_1=foreign_entity.bid, entity_2=foreign_key.primary_entity_bid)) # fmt: skip
        for candidate in self.primary_entity.attributes:
            if candidate.label.lstrip("#") == foreign_key.primary_key_label.lstrip("#"):
                self.primary_key = candidate
                break
        else:
            raise MocodoError(15, _('Attribute "{attribute_1}" in entity "{entity_1}" references an unknown attribute "{attribute_2}" in entity "{entity_2}".').format(attribute_1=foreign_key.label, entity_1=foreign_entity.bid, attribute_2=foreign_key.primary_key_label, entity_2=foreign_key.primary_entity_bid)) # fmt: skip

    def calculate_size(self, style, *ignored):
        self.fdx = self.foreign_entity.w // 2
        self.pdx = self.primary_entity.w // 2
        self.fdy = (
            -self.foreign_entity.h // 2
            + 3 * style["rect_margin_height"]
            + self.foreign_entity.cartouche_height
            + (self.foreign_key.rank + 0.5)
            * (self.foreign_entity.attribute_height + style["line_skip_height"])
        )
        self.pdy = (
            -self.primary_entity.h // 2
            + 3 * style["rect_margin_height"]
            + self.primary_entity.cartouche_height
            + (self.primary_key.rank + 0.5)
            * (self.primary_entity.attribute_height + style["line_skip_height"])
        )
        self.offset = 2 * (style["card_margin"] + style["card_max_width"])

    def description(self, style, geo):
        result = [("comment", {"text": f'Link from "{self.foreign_key.primary_key_label}" ({self.foreign_entity.bid}) to "{self.primary_key.label}" ({self.primary_entity.bid})'})]
        spins = (
            [(-1, -1), (1, -1), (-1, 1), (1, 1)]
            if self.foreign_key.rank % 2
            else [(1, 1), (-1, 1), (1, -1), (-1, -1)]
        )
        (fs, ps) = min(
            spins,
            key=lambda fs_ps: abs(
                geo["cx"][self.foreign_entity.bid]
                + self.fdx * fs_ps[0]
                - geo["cx"][self.primary_entity.bid]
                - self.pdx * fs_ps[1]
            ),
        )
        xf = geo["cx"][self.foreign_entity.bid] + self.fdx * fs
        yf = geo["cy"][self.foreign_entity.bid] + self.fdy
        xp = geo["cx"][self.primary_entity.bid] + self.pdx * ps
        yp = geo["cy"][self.primary_entity.bid] + self.pdy
        result.append(
            (
                "curve",
                {
                    "x0": xf,
                    "y0": yf,
                    "x1": xf + (xp - xf) / 2 if fs != ps else xf + self.offset * fs,
                    "y1": yf + (yp - yf) / 2,
                    "x2": xf + (xp - xf) / 3 if fs != ps else xp + self.offset * ps,
                    "y2": yp,
                    "x3": xp,
                    "y3": yp,
                    "stroke_color": style["leg_stroke_color"],
                    "stroke_depth": style["leg_stroke_depth"],
                }
            )
        )
        result.append(
            (
                "arrow",
                {
                    "x0": xp,
                    "y0": yp,
                    "x1": xp + ps * style["arrow_width"],
                    "y1": yp - ps * style["arrow_half_height"],
                    "x2": xp + ps * style["arrow_axis"],
                    "y2": yp,
                    "x3": xp + ps * style["arrow_width"],
                    "y3": yp + ps * style["arrow_half_height"],
                    "stroke_color": style["leg_stroke_color"],
                },
            ),
        )
        result.append(
            (
                "circle",
                {
                    "cx": xf,
                    "cy": yf,
                    "r": style["box_stroke_depth"],
                    "stroke_depth": style["box_stroke_depth"],
                    "stroke_color": style["leg_stroke_color"],
                    "color": style["leg_stroke_color"],
                }
            )
        )
        return result



import importlib

__import__("sys").path[0:0] = ["."]

from ..mocodo_error import subsubopt_error
from ..parse_mcd import Visitor
from ..tools.parser_tools import first_child, is_identifier
from ..tools.string_tools import rstrip_digit_or_underline
from ..tools.graphviz_tools import create_name_to_index


class Crow(Visitor):
    def __init__(self):
        self.name_to_index = create_name_to_index()
        self.tables = {}
        self.links = []
        self.has_no_datatype = True
        self.invisible_boxes = set()
    
    def entity_or_table_attr(self, tree):
        id_groups = str(first_child(tree, "id_groups"))
        id_mark = str(first_child(tree, "id_mark"))
        attr = str(first_child(tree, "attr"))
        datatype = str(first_child(tree, "datatype"))
        if datatype:
            self.has_no_datatype = False
        tree.children = [(id_groups, id_mark, attr, datatype)]
    
    def entity_clause(self, tree):
        ent_name = first_child(tree, "box_name").value
        if first_child(tree, "box_def_prefix") == "-":
            self.invisible_boxes.add(ent_name)
            return
        ent_index = self.name_to_index(ent_name)
        ent_name = rstrip_digit_or_underline(ent_name)
        attrs = []
        has_id = False
        for (i, node) in enumerate(tree.find_data("entity_or_table_attr")):
            (id_groups, id_mark, attr, datatype) = node.children[0]
            is_id = is_identifier(i, id_groups, id_mark)
            has_id = has_id or is_id
            attrs.append((datatype, attr, is_id))
        self.tables[ent_index] = (ent_name, has_id, attrs)
    
    def assoc_clause(self, tree):
        assoc_name = first_child(tree, "assoc_name_def").children[0].value
        if first_child(tree, "box_def_prefix") == "-":
            self.invisible_boxes.add(assoc_name)
            return
        cards = [node.children[0].value for node in tree.find_data("card")]
        assert len(cards) == 2
        assoc_name = rstrip_digit_or_underline(assoc_name)
        kind = ".." if any(s.children[0].value == "_" for s in tree.find_data("card_prefix")) else "--"
        entities = tree.find_data("entity_name_ref")
        ent_1 = first_child(next(entities), "entity_name_ref").children[0]
        ent_2 = first_child(next(entities), "entity_name_ref").children[0]
        self.links.append((ent_1, cards[1], kind, cards[0], ent_2, assoc_name))
    
    def start(self, tree):
        visible_links = []
        for (ent_1, card_1, kind, card_2, ent_2, assoc_name) in self.links:
            if ent_1 in self.invisible_boxes or ent_2 in self.invisible_boxes:
                continue
            visible_links.append((ent_1, card_1, kind, card_2, ent_2, assoc_name))


def run(source, subargs, common=None):
    extension = "mmd" if "mmd" in subargs or "mermaid" in subargs else "gv"
    try:
        module = importlib.import_module(f".convert.crow_{extension}", package="mocodo")
    except ModuleNotFoundError:
        raise subsubopt_error(extension)
    text = module.run(source, subargs, common)
    if extension == "mmd":
        return {
            "stem_suffix": "_erd_crow",
            "text": text,
            "extension": "mmd",
            "to_defer": True,
            "highlight": "mermaid",
        }
    else:
        return {
            "stem_suffix": "_erd_crow",
            "text": text,
            "extension": "gv",
            "to_defer": True,
            "highlight": "graphviz",
        }




from collections import defaultdict

__import__("sys").path[0:0] = ["."]

from ..parse_mcd import Token, Visitor
from ..tools.parser_tools import parse_source, reconstruct_source, first_child


class DrainerFirstPass(Visitor):
    """
    In the first pass, the attributes to drain are suppressed from the tree and
    stored in a dictionary.
    """
    def __init__(self):
        self.drained_attrs = defaultdict(list)

    def assoc_clause(self, tree):

        # Guard: ensure that exactly one cardinality is 11
        cards = [node.children[0].value for node in tree.find_data("card")]
        if cards.count("11") != 1:
            return

        # Guard: ensure that the association has at least one attributes
        attrs = [*tree.find_data("typed_attr")]
        if len(attrs) == 0:
            return

        # Find out the name of the entity that is the target of the 11 leg
        entity_name_refs = [node.children[0] for node in tree.find_data("entity_name_ref")]
        i = cards.index("11")
        entity_name_ref = entity_name_refs[i].children[0].value

        # Store the attributes to drain
        self.drained_attrs[entity_name_ref].extend(map(reconstruct_source, attrs))

        # Remove the attributes from the association
        drop = False
        children = []
        for child in tree.children:
            if isinstance(child, Token):
                if child.type == "COLON":
                    drop = True
                elif child.type == "NL":
                    drop = False
            if not drop:
                children.append(child)
        tree.children = children


class DrainerSecondPass(Visitor):
    """
    In the second pass, the drained attributes are reinjected into the appropriate entities.
    """
    def __init__(self, drained_attrs):
        self.drained_attrs = drained_attrs

    def entity_clause(self, tree):

        # Guard: ensure that the entity must get drained attributes
        entity_name_def = first_child(tree, "entity_name_def").children[0]
        if not entity_name_def in self.drained_attrs:
            return

        # Add the drained attributes to the entity
        attrs = self.drained_attrs[entity_name_def]
        last_token = tree.children[-1]
        last_token.value = f", {', '.join(attrs)}{last_token.value}"


def run(source, **kargs):
    tree = parse_source(source)
    visitor = DrainerFirstPass()
    visitor.visit(tree)
    visitor = DrainerSecondPass(visitor.drained_attrs)
    visitor.visit(tree)
    return reconstruct_source(tree)



from math import hypot

from .cross import cross


def fitness(links, multiplicity, col_count, row_count, max_distance = 4):
    """ Return (by closure) a function evaluating the aesthetic quality of a given layout. """
    
    def evaluate(layout):
        for (position, index) in enumerate(layout):
            coordinates[index] = divmod(position, col_count)
        segments = [(coordinates[p1], coordinates[p2], multiplicity[p1, p2]) for (p1, p2) in links]
        total_distances = 0
        short_segments = []
        for  ((y1, x1), (y2, x2), m) in segments:
            distance = distances[abs(x1-x2)][abs(y1-y2)] * m
            if distance <= max_distance:
                short_segments.append((x1, y1, x2, y2))
            total_distances += distance
        crossing_count = (link_count - len(short_segments)) * link_count
        for (i, (x1, y1, x2, y2)) in enumerate(short_segments):
             for (x3, y3, x4, y4) in short_segments[i+1:]:
                 crossing_count += cross(x1, y1, x2, y2, x3, y3, x4, y4)
        return (crossing_count, total_distances)
    
    distances = [[hypot(i, j) - 1 for j in range(row_count)] for i in range(col_count)]
    coordinates = [(0, 0)] * (row_count * col_count)
    link_count = len(links)
    return evaluate



import itertools
import gettext

__import__("sys").path[0:0] = ["."]

from ..parse_mcd import Visitor
from ..tools.parser_tools import first_child, parse_source, is_identifier
from ..rewrite import (
    _split as split,
    _drain as drain,
)
from ..version_number import version

MIN_MAX_IN_UML = {
    "01": "1..*",
    "0N": "*",
    "11": "1",
    "1N": "1..*",
}

gettext.NullTranslations().install()

INHERITANCE_IN_UML = {
    "XT": _("{complete, disjoint}"), # Partition
    "X": _("{incomplete, disjoint}"), # Exclusivité
    "T": _("{complete, overlapping}"), # Totalité
    "": _("{incomplete, overlapping}"),
}

class UmlClassDiagram(Visitor):
    def __init__(self, preamble, common):
        self.common = common
        self.preamble = preamble or "" # preamble may be None
        self.id_counter = itertools.count()
        self.df_label = common.params["df"]
        self.has_no_datatype = True
        self.acc = []
        self.invisible_boxes = set()

    def entity_or_table_attr(self, tree):
        id_groups = str(first_child(tree, "id_groups"))
        id_mark = str(first_child(tree, "id_mark"))
        attr = str(first_child(tree, "attr"))
        datatype = str(first_child(tree, "datatype"))
        if datatype:
            self.has_no_datatype = False
        tree.children = [(id_groups, id_mark, attr, datatype)]
    
    def entity_clause(self, tree):
        ent_name = str(first_child(tree, "box_name"))
        if first_child(tree, "box_def_prefix") == "-":
            self.invisible_boxes.add(ent_name)
            return
        attributes = []
        for (i, node) in enumerate(tree.find_data("entity_or_table_attr")):
            (id_groups, id_mark, attr_label, datatype) = node.children[0]
            attr_label = str(attr_label)
            if attr_label == "":
                continue # ignore spacer attributes
            attributes.append((is_identifier(i, id_groups, id_mark), attr_label, datatype))
        self.acc.append({
            "kind": "table",
            "name": ent_name,
            "attributes": attributes,
        })
        self.acc.append({"kind": "spacer"})
    
    def assoc_clause(self, tree):
        legs = [node for node in tree.find_data("assoc_leg")]
        assoc_name = str(first_child(tree, "assoc_name_def").children[0])
        if first_child(tree, "box_def_prefix") == "-":
            self.invisible_boxes.add(assoc_name)
            return

        min_maxs = [node.children[0].value for node in tree.find_data("card")]
        ent_names = [str(first_child(leg, "entity_name_ref").children[0]) for leg in legs]

        typed_attrs = []
        for node in list(tree.find_data("typed_attr")):
            attr_label = str(first_child(node, "attr"))
            if attr_label == "":
                continue # don't create a node for a spacer attribute
            datatype = str(first_child(node, "datatype"))
            typed_attrs.append((attr_label, datatype))

        if len(legs) == 2:
            weaks = [first_child(leg, "card_prefix") == "_" for leg in legs]
            self.acc.append({
                "kind": "binary_link",
                "tables": ent_names,
                "table_1": ent_names[0],
                "card_1": MIN_MAX_IN_UML.get(min_maxs[1], "*"),
                "composition_1": weaks[1],
                "table_2": ent_names[1],
                "card_2": MIN_MAX_IN_UML.get(min_maxs[0], "*"),
                "composition_2": weaks[0],
                "assoc_name": assoc_name,
            })
            if typed_attrs:
                self.acc.append({
                    "kind": "binary_link_attrs",
                    "tables": ent_names,
                    "table_1": ent_names[0],
                    "table_2": ent_names[1],
                    "assoc_name": assoc_name,
                })
        else:
            # The card is 1 (11) if the min-max card has a "/" prefix, 1..* or * otherwise
            diamond = f"N_ARY_{next(self.id_counter)}"
            self.acc.append({
                "kind": "n_ary_link",
                "diamond": diamond,
                "tables": ent_names,
                "cards": ["1" if first_child(leg, "card_prefix") == "/" else MIN_MAX_IN_UML.get(first_child(leg, "card"), "*") for leg in legs],
                "assoc_name": assoc_name,
            })
            if typed_attrs:
                self.acc.append({
                    "kind": "n_ary_link_attrs",
                    "diamond": diamond,
                    "tables": ent_names,
                    "assoc_name": assoc_name,
                })
        
        if typed_attrs:
            self.acc.append({
                "kind": "table",
                "name": assoc_name,
                "attributes": [(False, attr_label, datatype) for (attr_label, datatype) in typed_attrs],
                "tables": ent_names,
            })
        
        self.acc.append({"kind": "spacer"})
    
    def inheritance_clause(self, tree):
        inheritance_name = str(first_child(tree, "inheritance_name"))
        box_names = [str(first_child(leg, "box_name")) for leg in tree.find_data("box_name")]
        box_names.reverse() # not sure why, but the boxes seem to come in reverse order
        self.acc.append({
            "kind": "generalization",
            "set": INHERITANCE_IN_UML[inheritance_name],
            "parent": box_names[0],
            "children": [box_name for box_name in box_names[1:]],
            "id": f"GENERALIZATION_{next(self.id_counter)}",
        })
        self.acc.append({"kind": "spacer"})
    
    def start(self, tree):
        self.acc = list(filter(lambda d: not self.invisible_boxes.intersection(d.get("tables", [])), self.acc))
    
    def get_plantuml(self):
        style = self.common.load_style()
        result = []
        if not self.preamble.startswith("-"):
            result.append(f"' Generated by Mocodo {version}\n")
        result.append(f'@startuml "{self.common.params["title"]}"\n')
        result.append(f'!define Table(x) class "x" << (T,{style["entity_color"]}) >>')
        result.append("!define pk(x) <b>x</b>")
        if self.preamble.startswith("-"):
            self.preamble = self.preamble[1:]
        else:
            result.append("\n".join([
                "hide methods",
                "left to right direction",
                f'skinparam groupInheritance 2',
                f'skinparam lineThickness {style["box_stroke_depth"]}',
                f'skinparam lineColor {style["association_stroke_color"]}',
                f'skinparam backgroundColor {style["background_color"]}',
                f'skinparam classAttributeFontColor {style["entity_attribute_text_color"]}',
                f'skinparam classAttributeFontName Monospaced',
                f'skinparam classAttributeFontSize 14',
                f'skinparam classBackgroundColor {style["entity_color"]}',
                f'skinparam classBorderColor {style["entity_stroke_color"]}',
                f'skinparam classBorderThickness {style["box_stroke_depth"]}',
                f'skinparam classFontColor {style["entity_cartouche_text_color"]}',
                f'skinparam classFontName Arial',
                f'skinparam classFontSize 18',
                f'skinparam classHeaderBackgroundColor {style["entity_cartouche_color"]}',
            ]))
        result.append(self.preamble.replace(r"\n", "\n")) # the user-defined preamble comes raw
        for element in self.acc:
            if element["kind"] == "table":
                max_attr_length = max([len(attr) for (_, attr, _) in element["attributes"]], default=0)
                result.append(f'Table("{element["name"]}") {{')
                for (is_id, attr, datatype) in element["attributes"]:
                    suffix = ""
                    if datatype:
                        tabs = " " * (max_attr_length - len(attr) + 1)
                        suffix = f"{tabs}{datatype}"
                    if is_id:
                        attr = f"pk({attr})"
                    result.append(f"    {{field}} + {attr}{suffix}")
                result.append(f"}}")
            elif element["kind"] == "binary_link":
                head = "*" if element["composition_1"] else "-"
                tail = "*" if element["composition_2"] else "-"
                element["link"] = f"{head}-{tail}"
                if element["assoc_name"].upper() == self.df_label.upper():
                    result.append('"{table_1}" "{card_1}" {link} "{card_2}" "{table_2}"'.format(**element))
                else:
                    result.append('"{table_1}" "{card_1}" {link} "{card_2}" "{table_2}": "{assoc_name}"'.format(**element))
            elif element["kind"] == "n_ary_link":
                result.append(f"diamond {element['diamond']}")
                for (table_name, card) in zip(element["tables"], element["cards"]):
                    result.append(f'{element["diamond"]} -- "{card}" "{table_name}"')
            elif element["kind"] == "binary_link_attrs":
                result.append('("{table_1}", "{table_2}") .. "{assoc_name}"'.format(**element))
            elif element["kind"] == "n_ary_link_attrs":
                result.append('{diamond} "{assoc_name}" .. "{assoc_name}"'.format(**element))
            elif element["kind"] == "generalization":
                if element["set"]:
                    result.append(f'note "{element["set"]}" as {element["id"]}')
                    result.extend(f'{element["id"]} -[dotted]- {child}' for child in element["children"])
                result.extend(f'{element["parent"]} <|-- {child}' for child in element["children"])
            else: # spacer
                result.append("")
        result.append(f"@enduml\n")
        return "\n".join(result)


def run(source, subargs=None, common=None):
    (language, preamble) = next(iter(subargs.items()), ("plantuml", ""))
    source = split.run(source)
    source = drain.run(source)
    tree = parse_source(source)
    extractor = UmlClassDiagram(preamble, common)
    extractor.visit(tree)
    if True: # language == "plantuml":
        text = extractor.get_plantuml()
        extension = "puml"
        highlight = "puml"
    else:
        text = extractor.get_mermaid()
        extension = "mmd"
        highlight = "mermaid"
    return {
        "stem_suffix": "_uml",
        "text": text,
        "extension": extension,
        "to_defer": True,
        "highlight": highlight,
    }



from bisect import bisect_left
from pathlib import Path

from ..tools import load_mini_yaml
from ..mocodo_error import MocodoError

def read_template(stem_or_path, official_template_dir):

    def traverse_templates(stem_or_path, template_stack, already_seen):
        if stem_or_path.endswith(".yaml"):
            path = Path(stem_or_path).absolute()
        else:
            path = official_template_dir / f"{stem_or_path}.yaml"
        (folder, name) = (path.parent, path.stem)
        if stem_or_path in already_seen:
            raise MocodoError(30, _('Circular inheritance in template "{name}.yaml" of "{folder}.').format(name=name, folder=folder))  # fmt: skip
        already_seen.add(stem_or_path)
        if not path.is_file():
            raise MocodoError(31, _('Template "{name}.yaml" not found in "{folder}".').format(name=name, folder=folder))  # fmt: skip
        try:
            template = load_mini_yaml.run(path)
        except:
            raise MocodoError(32, _('Unable to decode template "{name}.yaml" of "{folder}".').format(name=name, folder=folder))  # fmt: skip
        for (key, array) in template.items():
            if isinstance(array, dict):
                raise MocodoError(34, _('Template "{name}.yaml" of "{folder}" contains a YAML object as value of key "{key}".').format(name=name, folder=folder, key=key))  # fmt: skip
            if not isinstance(array, list):
                continue
            previous_order = None
            for d in array:
                if not isinstance(d, dict):
                    raise MocodoError(35, _('Template "{name}.yaml" of "{folder}" contains a YAML array as value of key "{key}" which does not contain only YAML objects.').format(name=name, folder=official_template_dir, key=key))  # fmt: skip
                if "order" not in d:
                    raise MocodoError(36, _('Template "{name}.yaml" of "{folder}" contains a YAML array as value of key "{key}" which does not contain only objects having an "order" key.').format(name=name, folder=official_template_dir, key=key))  # fmt: skip
                order = d["order"]
                if not isinstance(order, (int, float)):
                    raise MocodoError(38, _('Template "{name}.yaml" of "{folder}" contains a YAML array as value of key "{key}" where the "order" key is not associated to a number.').format(name=name, folder=official_template_dir, key=key))
                if previous_order is not None and order <= previous_order:
                    raise MocodoError(39, _('Template "{name}.yaml" of "{folder}" contains a YAML array as value of key "{key}" where the "order" keys are not sorted in ascending order.').format(name=name, folder=official_template_dir, key=key))
                previous_order = order
        template_stack.append(template)
        if "parent" in template:
            return traverse_templates(template["parent"], template_stack, already_seen)
        else:
            return reversed(template_stack)
    
    result = {}
    for template in traverse_templates(stem_or_path, [], set()):
        for key in template:
            if not isinstance(result.get(key), list):
                # create or update a scalar value
                result[key] = template[key]
            else:
                if template[key] == []:
                    result[key] = []                    
                # update a non-empty list of dictionaries having an "order" key
                for new_dictionary in template[key]:
                    order = new_dictionary["order"]
                    orders = [d["order"] for d in result[key]] # Prior to Python 3.10, bisect_left has no `key` argument
                    i = bisect_left(orders, order)
                    if i < len(result[key]) and result[key][i]["order"] == order:
                        # a dictionary with the same order already exists
                        if set(new_dictionary.keys()).issubset({"order", "comment"}):
                            # the new dictionary defines nothing: remove the existing dictionary
                            del result[key][i]
                        else:
                            # update the existing dictionary in place
                            result[key][i].update(new_dictionary)
                    else:
                        # insert the new dictionary at the right place
                        result[key].insert(i, new_dictionary)
    result.pop("parent", None)
    return result



import textwrap
import re
from unicodedata import combining, normalize
import base64
import zlib

TRUNCATE_DEFAULT_SIZE = 64

def wrap_label(s, wrapping_ratio=2):
    """
    Break a given short string into a list of lines, trying to keep the
    width/height ratio just above wrapping_ratio.
    """
    min_width = max(map(len, re.split(r"[-\s]+", s)))
    for width in range(min_width, len(s)):
        lines = textwrap.wrap(s, width=width)
        if max(map(len, lines)) / len(lines) > wrapping_ratio:
            return lines
    return [s]

def rstrip_digit_or_underline(s):
    """
    Get rid of single digit suffix, if any. Works on empty strings too.
    Used for entity and association raw names.
    """
    return s[:-1] if s[-1:].isdigit() or s[-1:] == "_" else s

def surrounds(s, ends):
    """
    Check if s is surrounded by the first and last characters of ends.
    Works on empty strings too.
    """
    return s[:1] == ends[:1] and s[-1:] == ends[-1:]

def strip_surrounds(s, ends):
    """
    Remove the first and last characters of s if they match the first and last
    characters of ends. Works on empty strings too.
    """
    return s[1:-1] if surrounds(s, ends) else s

def is_a_description(note):
    if not note:
        return False
    if note.startswith(("+", "-")):
        return False
    return " " in note

# ASCII

LATIN = "ä  æ  ǽ  đ ð ƒ ħ ı ł ø ǿ ö  œ  ß  ŧ ü  Ä  Æ  Ǽ  Đ Ð Ƒ Ħ I Ł Ø Ǿ Ö  Œ  ẞ  Ŧ Ü "
ASCII = "ae ae ae d d f h i l o o oe oe ss t ue AE AE AE D D F H I L O O OE OE SS T UE"


def ascii(s, outliers=str.maketrans(dict(zip(LATIN.split(), ASCII.split())))):
    return "".join(c for c in normalize("NFD", s.translate(outliers)) if not combining(c))

def raw_to_bid(box_raw_name):
    return re.sub(r"\W", "_", ascii(box_raw_name).upper())

def aggressive_split(
    input_string,
    find_all_words = re.compile(r"(?u)[^\W_]+").findall
):
    # Split the string, and further split the words
    result = []
    for word in find_all_words(input_string):
        previous_is_digit = word[0].isdigit()
        previous_is_lower = word[0].islower()
        i = 0
        for (j, c) in enumerate(word[1:], 1):
            current_is_lower = c.islower()
            current_is_digit = c.isdigit()
            if (
                previous_is_digit and not current_is_digit
                or
                not previous_is_digit and current_is_digit
                or
                previous_is_lower and not current_is_lower
            ):
                result.append(word[i:j])
                i = j
            previous_is_lower = current_is_lower
            previous_is_digit = current_is_digit
        result.append(word[i:])
    return result

# A decorator which calls

def spare_protected_suffix(
    function,
    is_protected_suffix = re.compile(r"[_0-9]").match,
):
    def wrapper(input_string):
        suffix = input_string[-1:]
        if is_protected_suffix(suffix):
            input_string = input_string[:-1]
        else:
            suffix = ""
        result = function(input_string)
        return result + suffix
    return wrapper

@spare_protected_suffix
def camel(input_string):
    result = "".join([word.capitalize() for word in aggressive_split(input_string)])
    return result[:1].lower() + result[1:]

@spare_protected_suffix
def pascal(input_string):
    return "".join([word.capitalize() for word in aggressive_split(input_string)])

@spare_protected_suffix
def snake(input_string):
    return "_".join([word for word in aggressive_split(input_string)])

def urlsafe_encoding(text):
    return base64.urlsafe_b64encode(zlib.compress(text.encode('utf-8'), 9)).decode('ascii')

# Markdown

def markdown_table(rows, formats=None):
    """
    Return a Markdown table from a list of rows, each row being a tuple.
    The first row is the header. For maximum readability, in text mode, each column
    is filled with spaces so that the longest cell fits in the column.
    'formats' is a sequence of formatting symbols "l", "c", "r".
    """
    rows = [list(map(str, row)) for row in rows]
    widths = [max(map(len, column)) for column in zip(*rows)]
    result = ["| " + " | ".join(f"{c:<{w}}" for (w, c) in zip(widths, row)) + " |" for row in rows]
    formats = formats or "l" * len(rows[0])
    formats = [("-" if x == "r" else ":") + "-" * w + ("-" if x == "l" else ":") for (w, x) in zip(widths, formats)]
    result[1:1] = ["|" + "|".join(formats) + "|"]
    return "\n".join(result)



from .tools.string_tools import raw_to_bid

class Attribute:

    # The following three static attributes will be updated by Mcd.add_attributes() method.
    # They are here to simplify the writing of some tests.
    id_gutter_strong_string = "ID"
    id_gutter_weak_string = "id"
    id_gutter_alts = dict(zip("123456789", "123456789"))

    def __init__(self, attribute):
        self.label = attribute.get("attribute_label", "")
        self.label_view = self.label # may be updated in self.register_foreign_key_status()
        self.rank = attribute["rank"]
        self.datatype = attribute.get("datatype", "")
        self.primary_entity_bid = raw_to_bid(attribute.get("that_table", ""))
        self.id_groups = set(attribute.get("id_groups", "").replace("0", ""))
        self.id_text =  " ".join(map(self.id_gutter_alts.get, sorted(self.id_groups)))
        self.id_gutter_width = 0  # For anything but entities

    def register_foreign_key_status(self, attribute, fk_format):
        that_table = attribute.get("that_table")
        if not that_table:
            return
        self.primary_key_label = attribute.get("that_table_attribute_label")
        self.label_view = fk_format.format(label=self.label)

    def calculate_size(self, style, get_font_metrics):
        self.attribute_font = style[self.font_type]
        self.font = get_font_metrics(self.attribute_font)
        self.w = self.font.get_pixel_width(self.label_view)
        self.h = self.font.get_pixel_height()
        self.id_width = self.font.get_pixel_width(self.id_text)

    def set_id_gutter_width(self, id_gutter_width):
        self.id_gutter_width = id_gutter_width

    def description(self, style, x, y, dx, dy):
        result = [
            (
                "text",
                {
                    "x": x + dx,
                    "y": y + round(dy + style["attribute_text_height_ratio"] * self.h, 1),
                    "text": self.label_view,
                    "text_color": style[f"{self.box_type}_attribute_text_color"],
                    "family": self.attribute_font["family"],
                    "size": self.attribute_font["size"],
                }
            )
        ]
        if self.id_gutter_width:
            result.append(
                (
                    "text",
                    {
                        "x": x + (self.id_gutter_width - self.id_width) // 2 - style["rect_margin_width"],
                        "y": y + round(dy + style["attribute_text_height_ratio"] * self.h, 1),
                        "text": self.id_text,
                        "text_color": style[f"{self.box_type}_attribute_text_color"],
                        "family": self.attribute_font["family"],
                        "size": self.attribute_font["size"],
                    }
                )
            )
        return result

class SimpleEntityAttribute(Attribute):

    def __init__(self, attribute):
        Attribute.__init__(self, attribute)
        self.box_type = "entity"
        self.font_type = "entity_attribute_font"
        self.kind = "simple"


class SimpleAssociationAttribute(Attribute):

    def __init__(self, attribute):
        Attribute.__init__(self, attribute)
        self.box_type = "association"
        self.font_type = "association_attribute_font"
        self.kind = "simple_association"


class IdentifierAttribute(Attribute):

    def __init__(self, attribute):
        Attribute.__init__(self, attribute)
        self.box_type = "entity"
        self.font_type = "entity_attribute_font"
        self.id_groups.add("0")

    def calculate_size(self, *args):
        Attribute.calculate_size(self, *args)

class StrongAttribute(IdentifierAttribute):

    def __init__(self, attribute):
        IdentifierAttribute.__init__(self, attribute)
        self.id_text = f"{self.id_text} {self.id_gutter_strong_string}".lstrip()
        self.kind = "strong"

    def description(self, style, x, y, dx, dy):
        return IdentifierAttribute.description(self, style, x, y, dx, dy) + [
            (
                "line",
                {
                    "x0": x + dx,
                    "y0": y + dy + self.h + style["underline_skip_height"],
                    "x1": x + dx + self.w,
                    "y1": y + dy + self.h + style["underline_skip_height"],
                    "stroke_depth": style["underline_depth"],
                    "stroke_color": style['entity_attribute_text_color'],
                }
            )
        ]

class WeakAttribute(IdentifierAttribute):

    def __init__(self, attribute):
        IdentifierAttribute.__init__(self, attribute)
        self.id_text = f"{self.id_text} {self.id_gutter_weak_string}".lstrip()
        self.kind = "weak"

    def description(self, style, x, y, dx, dy):
        return IdentifierAttribute.description(self, style, x, y, dx, dy) + [
            (
                "dash_line",
                {
                    "x0": x + dx,
                    "y0": y + dy + self.h + style["underline_skip_height"],
                    "x1": x + dx + self.w,
                    "y1": y + dy + self.h + style["underline_skip_height"],
                    "dash_width": style["dash_width"],
                    "stroke_depth": style["underline_depth"],
                    "stroke_color": style['entity_attribute_text_color'],
                }
            )
        ]


class PhantomAttribute(Attribute):
    
    def __init__(self, attribute):
        Attribute.__init__(self, attribute)
        self.kind = "phantom"
        self.font_type = "entity_attribute_font" # dummy
    
    def description(self, *_):
        return []


class InheritanceAttribute(Attribute):

    def __init__(self, attribute):
        Attribute.__init__(self, attribute)
        self.kind = "inheritance"



import importlib
from time import time
from mocodo.mocodo_error import subarg_error

def run(source, subargs, **kargs):
    algo = subargs.get("algo", "bb")
    try:
        module = importlib.import_module(f".rewrite.arrange_{algo}", package="mocodo")
    except ModuleNotFoundError:
        raise subarg_error("algo", algo)
    timeout = subargs.get("timeout")
    if timeout is None:
        has_expired = lambda: False
    else:
        try:
            timeout = time() + float(timeout)
            has_expired = lambda: time() > timeout
        except Exception:
            raise subarg_error("timeout", timeout)
    return module.arrange(source, subargs, has_expired)



from ..parse_mcd import Visitor
from ..tools.parser_tools import first_child, parse_source, reconstruct_source

class ArrowsHere(Visitor):

    def assoc_leg(self, tree):
        card = first_child(tree, "card")
        if card != "11":
            return
        if first_child(tree, "leg_arrow") != "":
            return
        card.value += "<"
    
class ArrowsAcross(Visitor):

    def assoc_clause(self, tree):
        legs = [node for node in tree.find_data("assoc_leg")]
        cards = [first_child(leg, "card") for leg in legs]
        for (i, card) in enumerate(cards):
            if card == "11":
                for (j, other_leg) in enumerate(legs):
                    if i != j and first_child(other_leg, "leg_arrow") == "":
                        other_card = first_child(other_leg, "card")
                        other_card.value += ">"


def create_df_arrows(source, where):
    visitor = ArrowsHere() if where == "here" else ArrowsAcross()
    tree = parse_source(source)
    visitor.visit(tree)
    return reconstruct_source(tree)



from collections import Counter
import itertools
import re

__import__("sys").path[0:0] = ["."]

from ..parse_mcd import Visitor
from ..tools.parser_tools import first_child, parse_source, is_identifier
from ..rewrite import _split as split
from ..tools.graphviz_tools import create_name_to_index
from ..tools.string_tools import wrap_label, rstrip_digit_or_underline
from ..version_number import version

class Chen(Visitor):
    def __init__(self, subargs, common):
        self.no_attrs = "attrs" not in subargs
        subargs.pop("attrs", None)
        self.subargs = subargs
        self.df_counter = itertools.count()
        self.df_label = common.params["df"]

        self.name_to_index = create_name_to_index()
        # Entities and their attributes
        self.entity_nodes = []
        self.weak_box_indexes = set()
        self.strengthening_leg_count = Counter()
        self.ent_attr_nodes = []
        self.ent_key_attr_nodes = []
        self.ent_attr_edges = []
        self.not_gerund_nodes = set()
        # Associations and their attributes
        self.rel_nodes = []
        self.rel_attr_nodes = []
        self.rel_attr_edges = []
        # Legs with their participation
        self.partial_edges = []
        self.total_edges = []
        # Invisible boxes
        self.invisible_boxes = set()

    def entity_or_table_attr(self, tree):
        id_groups = str(first_child(tree, "id_groups"))
        id_mark = str(first_child(tree, "id_mark"))
        attr = str(first_child(tree, "attr"))
        tree.children = [(id_groups, id_mark, attr)]
    
    def entity_clause(self, tree):
        ent_name = str(first_child(tree, "box_name"))
        ent_index = self.name_to_index(f"ent_{ent_name}")
        if first_child(tree, "box_def_prefix") == "-":
            self.invisible_boxes.add(ent_index)
            return
        ent_name = rstrip_digit_or_underline(ent_name)
        self.entity_nodes.append((ent_index, ent_name))

        # Entity attributes
        if self.no_attrs:
            return
        for (i, node) in enumerate(tree.find_data("entity_or_table_attr")):
            (id_groups, id_mark, attr_label) = node.children[0]
            attr_label = str(attr_label)
            if attr_label == "":
                continue # don't create a node for a spacer attribute
            attr_index = self.name_to_index(f"ent_attr_{ent_index}_{attr_label}")
            if is_identifier(i, id_groups, id_mark):
                self.ent_key_attr_nodes.append((attr_index, attr_label, ent_index))
            else:
                self.ent_attr_nodes.append((attr_index, attr_label))
            self.ent_attr_edges.append((ent_index, attr_index))
    
    def assoc_clause(self, tree):
        legs = [node for node in tree.find_data("assoc_leg")]

        # Accumulate the association as a new relationship
        assoc_name = str(first_child(tree, "assoc_name_def").children[0])
        if assoc_name == self.df_label:
            assoc_name = f"{assoc_name}{next(self.df_counter)}"
        assoc_index = self.name_to_index(f"assoc_{assoc_name}")
        if first_child(tree, "box_def_prefix") == "-":
            self.invisible_boxes.add(assoc_index)
            return
        assoc_name = rstrip_digit_or_underline(assoc_name)
        self.rel_nodes.append((assoc_index, assoc_name))

        min_maxs = [node.children[0].value for node in tree.find_data("card")]
        ent_names = [str(first_child(leg, "entity_name_ref").children[0]) for leg in legs]

        edge_accs = [self.partial_edges if c[0] == "0" else self.total_edges for c in min_maxs]
        ent_indexes = [self.name_to_index(f"ent_{ent_name}") for ent_name in ent_names]

        if len(legs) == 2:
            weaks = [first_child(leg, "card_prefix") == "_" for leg in legs]
            if any(weaks): # the binary association is "weak", in the Chen sense
                self.weak_box_indexes.add(assoc_index)
            # If both legs are *N, change one with *M
            if min_maxs[0][1] == min_maxs[1][1] == "N":
                min_maxs[0] = f"{min_maxs[0][0]}M"
            for i in range(2):
                # The 1-i is to get the other leg, i.e., to look across instead of to look here
                edge_accs[i].append((ent_indexes[i], assoc_index, min_maxs[1-i][1]))
                if weaks[i]: # the entity is distinguished by a _11
                    self.weak_box_indexes.add(ent_indexes[i])
                    self.strengthening_leg_count[ent_indexes[i]] += 1
                else: # we are sure that it is not an associative entity
                    self.not_gerund_nodes.add(ent_indexes[i])
        else:
            # The Chen's card is 1 if the min-max card has a "/" prefix, N otherwise
            chen_cards = ["1" if first_child(leg, "card_prefix") == "/" else "N" for leg in legs]
            for (chen_card, edge_acc, ent_index) in zip(chen_cards, edge_accs, ent_indexes):
                edge_acc.append((ent_index, assoc_index, chen_card))
                self.not_gerund_nodes.add(ent_index)
        
        # Association attributes
        if self.no_attrs:
            return
        for node in tree.find_data("attr"):
            attr_label = str(node.children[0])
            if attr_label == "":
                continue # don't create a node for a spacer attribute
            attr_index = self.name_to_index(f"assoc_attr_{assoc_index}_{attr_label}")
            self.rel_attr_nodes.append((attr_index, attr_label))
            self.rel_attr_edges.append((assoc_index, attr_index))
    
    def start(self, tree):
        # Since the traversal is depth-first, this is in fact a post-treatment
        for (ent_index, strengthening_leg_count) in self.strengthening_leg_count.items():
            if strengthening_leg_count < 2:
                self.not_gerund_nodes.add(ent_index)
        self.partial_edges = [(x, y, c) for (x, y, c) in self.partial_edges if x not in self.invisible_boxes and y not in self.invisible_boxes]
        self.total_edges = [(x, y, c) for (x, y, c) in self.total_edges if x not in self.invisible_boxes and y not in self.invisible_boxes]
    
    def get_graphviz(self, common):
        style = common.load_style()
        result = []
        result.append(f'// Generated by Mocodo {version}\n')
        result.append(f'graph{{')
        result.append(f'  bgcolor="{style["background_color"]}"')
        if common.params["seed"] is not None:
            result.append(f'  start={common.params["seed"]}')

        for (subsubopt, subsubarg) in self.subargs.items():
            result.append(f'  {subsubopt}={subsubarg}')

        result.append(f'\n  // Entities')
        result.append(f'  node [')
        result.append(f'    shape=box')
        result.append(f'    style=filled')
        result.append(f'    penwidth={style["box_stroke_depth"]}')
        result.append(f'    color="{style["entity_stroke_color"]}"')
        result.append(f'    fillcolor="{style["entity_cartouche_color"]}"')
        result.append(f'    fontcolor="{style["entity_cartouche_text_color"]}"')
        result.append(f'  ]')
        for (index, name) in self.entity_nodes:
            name = "\\n".join(wrap_label(name))
            if index in self.not_gerund_nodes:
                if index in self.weak_box_indexes:
                    result.append(f'  {index} [label="{name}",peripheries=2]')
                else:
                    result.append(f'  {index} [label="{name}"]')
        
        if len(self.entity_nodes) > len(self.not_gerund_nodes):
            result.append(f'\n  // Associative entities')
            for (index, name) in self.entity_nodes:
                name = "\\n".join(wrap_label(name))
                if index not in self.not_gerund_nodes:
                    result.append(f'  {index} [label="{name}",shape=Mdiamond]')

        if self.ent_attr_nodes or self.ent_key_attr_nodes:
            result.append(f'\n  // Normal entity attributes')
            result.append(f'  node [')
            result.append(f'    shape=oval')
            result.append(f'    penwidth={style["box_stroke_depth"]}')
            result.append(f'    color="{style["entity_stroke_color"]}"')
            result.append(f'    fillcolor="{style["entity_color"]}"')
            result.append(f'    fontcolor="{style["entity_attribute_text_color"]}"')
            result.append(f'  ]')
            for (index, name) in self.ent_attr_nodes:
                name = "\\n".join(wrap_label(name))
                result.append(f'  {index} [label="{name}"]')
            result.append(f'\n  // Weak and strong entity attributes')
            for (index, name, ent_index) in self.ent_key_attr_nodes:
                name = "<br/>".join(wrap_label(name))
                if ent_index in self.weak_box_indexes:
                    result.append(f'  {index} [label=<<u>{name}</u>> style="dashed,filled"]')
                else:
                    result.append(f'  {index} [label=<<u>{name}</u>>]')
        
        if self.rel_attr_nodes:
            result.append(f'\n  // Relationship attributes')
            result.append(f'  node [')
            result.append(f'    color="{style["association_stroke_color"]}"')
            result.append(f'    fillcolor="{style["association_color"]}"')
            result.append(f'    fontcolor="{style["association_attribute_text_color"]}"')
            result.append(f'  ]')
            for (index, name) in self.rel_attr_nodes:
                name = "\\n".join(wrap_label(name))
                result.append(f'  {index} [label="{name}"]')

        if self.rel_nodes:
            result.append(f'\n  // Relationships')
            result.append(f'  node [')
            result.append(f'    shape=diamond')
            result.append(f'    height=0.7')
            result.append(f'    penwidth={style["box_stroke_depth"]}')
            result.append(f'    color="{style["association_stroke_color"]}"')
            result.append(f'    fillcolor="{style["association_cartouche_color"]}"')
            result.append(f'    fontcolor="{style["association_cartouche_text_color"]}"')
            result.append(f'  ]')
            for (index, name) in self.rel_nodes:
                name = "\\n".join(wrap_label(name))
                if index in self.weak_box_indexes:
                    result.append(f'  {index} [label="{name}",peripheries=2]')
                else:
                    result.append(f'  {index} [label="{name}"]')

        if self.ent_attr_edges or self.rel_attr_edges:
            result.append(f'\n  // Edges between entities and attributes')
            result.append(f'  edge [')
            result.append(f'    penwidth={style["box_stroke_depth"]}')
            result.append(f'    color="{style["entity_stroke_color"]}"')
            result.append(f'  ]')
            for (ent_index, attr_index) in sorted(self.ent_attr_edges):
                result.append(f'  {ent_index} -- {attr_index}')
            result.append(f'\n  // Edges between relationships and attributes')
            result.append(f'  edge [color="{style["association_stroke_color"]}"]')
            useless = True
            for (assoc_index, attr_index) in sorted(self.rel_attr_edges):
                result.append(f'  {assoc_index} -- {attr_index}')
                useless = False
            if useless:
                result.pop()
                result.pop()

        if self.partial_edges or self.total_edges:
            seen_edges = set()
            result.append(f'\n  // Edges between entities and relationships')
            result.append(f'  edge [')
            result.append(f'    penwidth={style["leg_stroke_depth"]}')
            result.append(f'    color="{style["leg_stroke_color"]}:{style["leg_stroke_color"]}"') # default to total participation
            result.append(f'    fontcolor="{style["card_text_color"]}"')
            result.append(f'    labeldistance=2')
            result.append(f'    headlabel=1')
            result.append(f'  ]')
            for label in "1MN":
                useless = label != "1"
                if label != "1":
                    result.append(f'  edge [headlabel={label}]')
                suffix = ""
                for edges in (self.total_edges, self.partial_edges):
                    for (ent_index, assoc_index, edge_label) in sorted(edges):
                        if label == edge_label:
                            if (ent_index, assoc_index) in seen_edges:
                                (assoc_index, ent_index) = (ent_index, assoc_index)
                            result.append(f'  {ent_index} -- {assoc_index}{suffix}')
                            seen_edges.add((ent_index, assoc_index))
                            useless = False
                    suffix = f' [color="{style["leg_stroke_color"]}"]' # from now on, partial participation
                if useless:
                    result.pop()
        result.append("}")

        result = "\n".join(result)
        useless_colors = [
            'bgcolor="#FFFFFF"',
            'color="#000000"',
            'fontcolor="#000000"',
        ]
        for useless_color in useless_colors:
            result = re.sub(fr'(?m)^ *{useless_color}\n', "", result)
        return result

    def get_text_for_testing(self):
        nodes = {}
        for (i, x) in self.entity_nodes:
            if i in self.not_gerund_nodes:
                if i in self.weak_box_indexes:
                    nodes[i] = f"[[{x}]]"
                else:
                    nodes[i] = f"[{x}]"
            else:
                nodes[i] = f"[<{x}>]"
        for (i, x) in self.rel_nodes:
            if i in self.weak_box_indexes:
                nodes[i] = f"<<{x}>>"
            else:
                nodes[i] = f"<{x}>"
        for (i, x, j) in self.ent_key_attr_nodes:
            if j in self.weak_box_indexes:
                nodes[i] = f"(.{x}.)"
            else:
                nodes[i] = f"(_{x}_)"
        nodes.update((i, f"({x})") for (i, x) in self.ent_attr_nodes + self.rel_attr_nodes)
        result = []
        for (x, y, c) in self.partial_edges:
            result.append(f"            {nodes[x]} --{c}-- {nodes[y]}")
        for (x, y, c) in self.total_edges:
            result.append(f"            {nodes[x]} =={c}== {nodes[y]}")
        for (x, y) in self.ent_attr_edges + self.rel_attr_edges:
            result.append(f"            {nodes[x]} -- {nodes[y]}")
        return "\n".join(sorted(result))


def run(source, subargs=None, common=None, testing=False):
    subargs = subargs or {}
    source = split.run(source)
    tree = parse_source(source)
    extractor = Chen(subargs, common)
    extractor.visit(tree)
    if testing:
        return extractor.get_text_for_testing()
    text = extractor.get_graphviz(common)
    return {
        "stem_suffix": "_erd_chen",
        "text": text,
        "extension": "gv",
        "to_defer": True,
        "highlight": "graphviz",
    }



from pathlib import Path
from json import loads

def repr_single(s):
    """repr() but with single quotes"""
    return "'" + repr('"' + s)[2:]

def convert(source_path, target_path):
    for path in sorted(source_path.glob("*.json"), reverse=True):
        print(path)
        data = loads(path.read_text(encoding="utf8"))
        result = []
        for (key, stuff) in data.items():
            if isinstance(stuff, str):
                stuff = repr_single(stuff).replace(r"\\", "\\").replace("\\'", "''")
                result.append(f"{key}: {stuff}")
            elif isinstance(stuff, list):
                result.append(f"{key}:")
                for d in stuff:
                    sub_result = []
                    for (k, v) in d.items():
                        if isinstance(v, str):
                            v = repr_single(v).replace(r"\\", "\\").replace("\\'", "''")
                        else:
                            v = str(v)
                        sub_result.append(f"{k}: {v}")
                    result.append("  - " + "\n    ".join(sub_result))
            elif isinstance(stuff, bool):
                result.append(f"{key}: {str(stuff).lower()}")
            elif stuff is None:
                result.append(f"{key}: null")
            else:
                result.append(f"{key}: {stuff}")
        result.append("")
        dest = Path(target_path, path.stem + ".yaml")
        dest.write_text("\n".join(result), encoding="utf8")

if __name__ == "__main__":
    convert(Path("test/test_data/templates"), Path("test/test_data/new_templates"))
